from PySide6.QtWidgets import (
    QDialog,
    QDialogButtonBox,
    QVBoxLayout,
    QLabel
)


def open_error_dialog(error, buttons = QDialogButtonBox.Ok, parent=None):
    dialog = ErrorDialog(str(error), buttons, parent=parent)
    dialog.show()
    return dialog


class ErrorDialog(QDialog):
    Retry = QDialogButtonBox.Retry
    Cancel = QDialogButtonBox.Cancel

    def __init__(self, message, buttons, parent=None):
        super().__init__(parent=parent)

        self.setWindowTitle("Error")

        self.buttonBox = QDialogButtonBox(buttons)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

        # Set default button
        if self.buttonBox.button(QDialogButtonBox.Retry) is not None:
            self.buttonBox.button(QDialogButtonBox.Retry).setDefault(True)
        elif self.buttonBox.button(QDialogButtonBox.Ok) is not None:
            self.buttonBox.button(QDialogButtonBox.Ok).setDefault(True)

        self.layout = QVBoxLayout()
        display_text = QLabel(message)
        self.layout.addWidget(display_text)
        self.layout.addWidget(self.buttonBox)
        self.setLayout(self.layout)



# General utility imports
from icarus_v2.backend.logger import Logger
from PySide6.QtCore import Signal, QThread
from time import sleep
from threading import Lock
from icarus_v2.utils.udev_setup import setup_udev_rules
from icarus_v2.utils.raw_log_reader import RawLogReader
from pathlib import Path
# Data collection & Device imports
from icarus_v2.backend.dataq_interface import DataqInterface
from icarus_v2.backend.buffer_loader import BufferLoader
from icarus_v2.backend.pulse_generator import PulseGenerator
from icarus_v2.backend.event import Event
# Event handler imports
from icarus_v2.backend.pressurize_handler import PressurizeHandler
from icarus_v2.backend.depressurize_handler import DepressurizeHandler
from icarus_v2.backend.period_handler import PeriodHandler
from icarus_v2.backend.pressure_handler import PressureHandler
from icarus_v2.backend.pump_handler import PumpHandler
from icarus_v2.backend.log_handler import LogHandler
from icarus_v2.backend.sentry import Sentry
from icarus_v2.backend.sample_sensor_detector import SampleSensorDetector


# Define the DataHandler class
class DataHandler(QThread):
    # Send events to gui
    pressurize_event_signal = Signal(Event)
    depressurize_event_signal = Signal(Event)
    period_event_signal = Signal(Event)
    pressure_event_signal = Signal(Event)
    pump_event_signal = Signal(Event)
    # Tell gui when device is connected
    acquiring_signal = Signal(bool)
    # display error dialog
    display_error = Signal(str)
    # Show warning or error in toolbar
    toolbar_warning = Signal(str,str)
    # Signal to tell device control panel to shut down
    # Ideally this is unnecessary as the signal should be directly sent to the pulse_generator
    # But currently the control panel can not check the state of the device. This is a workaround.
    shutdown_signal = Signal()
    # Start new log file
    log_signal = Signal(bool)
    # Tell GUI whether the sample sensor is connected
    sample_sensor_connected = Signal(bool)

    def __init__(self):
        super().__init__()

        self.pulse_generator = PulseGenerator()
        self.connecting = False
        self.connected = False
        self.device = None
        self.logger = None

        # TESTING ONLY!!!. reads a raw data file instead of connecting to a device
        # edit raw_file to change the playback to a different file
        self.load_raw = False
        if self.load_raw:
            raw_file = "2.1_to_1.5kBar.xz"
            project_root = Path(__file__).resolve().parents[3]
            self.raw_file = project_root / 'logs' / 'raw' / raw_file

        # Loads data from device into buffer
        self.loader = BufferLoader()
        self.loader.device_disconnected.connect(self.device_disconnected)

        # Event handlers
        sample_rate = 4000
        event_update_hz = 30
        pressure_update_hz = 5
        event_display_bounds = (-10, 140)
        self.pressurize_handler = PressurizeHandler(
            self.loader,
            self.pressurize_event_signal,
            sample_rate,
            event_update_hz,
            event_display_bounds
        )
        self.depressurize_handler = DepressurizeHandler(
            self.loader,
            self.depressurize_event_signal,
            sample_rate,
            event_update_hz,
            event_display_bounds
        )
        self.period_handler = PeriodHandler(
            self.loader,
            self.period_event_signal,
            sample_rate,
            event_update_hz,
            event_display_bounds
        )
        self.pressure_handler = PressureHandler(
            self.loader,
            self.pressure_event_signal,
            sample_rate,
            pressure_update_hz)
        self.pump_handler = PumpHandler(self.loader, self.pump_event_signal, sample_rate)
        self.log_handler = LogHandler(self.loader, self.log_signal, sample_rate, pressure_update_hz)

        # Sentry
        self.sentry = Sentry()
        self.log_signal.connect(self.sentry.handle_experiment)
        self.pump_event_signal.connect(self.sentry.handle_pump)
        self.depressurize_event_signal.connect(self.sentry.handle_depressurize)
        self.sentry.warning_signal.connect(lambda x: self.toolbar_warning.emit(str(x),"orange"))
        self.sentry.error_signal.connect(lambda x: self.toolbar_warning.emit(str(x),"red")) 
        self.sentry.error_signal.connect(lambda x: self.display_error.emit(str(x))) 
        self.sentry.error_signal.connect(lambda x: self.shutdown_signal.emit())

        # Logger
        if not self.load_raw:
            self.logger = Logger()
            self.pressurize_event_signal.connect(self.logger.log_event)
            self.depressurize_event_signal.connect(self.logger.log_event)
            self.period_event_signal.connect(self.logger.log_event)
            self.pressure_event_signal.connect(self.logger.log_event)
            self.log_signal.connect(self.logger.new_log_file)

            # self.display_error.connect(self.logger.log_error)
            # self.toolbar_warning.connect(self.logger.log_error)

            self.sentry.warning_signal.connect(self.logger.log_error)
            self.sentry.error_signal.connect(self.logger.log_error)

        # Sample sensor detector
        self.sample_sensor_detector = SampleSensorDetector(self.sample_sensor_connected)
        self.depressurize_event_signal.connect(self.sample_sensor_detector.detect)

        # Prevents another thread from quitting this while it is initializing the device.
        # Without this, the cleanup code would run and then the QThreads would be initialized.
        self.quit_lock = Lock()

        self.acquiring_signal.emit(False)

    # Connect to a device
    def run(self):
        self.connecting = True
        self.quit_lock.acquire()
        udev_installed = False

        while self.connecting:

            # TESTING ONLY
            if self.load_raw:
                self.device = RawLogReader(self.raw_file)
                self.connecting = False
                self.connected = True
                break

            try:
                self.device = DataqInterface()
                self.connecting = False
                self.connected = True
            except Exception as e:
                # Continue connecting
                if (
                        "USB device not found" in str(e) or
                        "No such device (it may have been disconnected)" in str(e) or
                        "Operation timed out" in str(e) or
                        "Input/Output Error" in str(e) or
                        "No backend available" in str(e)
                        ):
                    self.quit_lock.release()
                    sleep(0.5)
                    if not self.connecting:
                        return
                    self.quit_lock.acquire()
                elif "Insufficient permissions to access the USB device" in str(e):
                    try:
                        if not udev_installed:
                            setup_udev_rules()
                            udev_installed = True
                        sleep(1)
                    except Exception as err:
                        self.display_error.emit(str(err))
                        break
                else:
                    self.display_error.emit(str(e))
                    break

        if self.device is not None:
            self.loader.set_device(self.device)
            self.pulse_generator.set_device(self.device)

            self.pressurize_handler.start()
            self.depressurize_handler.start()
            self.period_handler.start()
            self.pressure_handler.start()
            self.pump_handler.start()
            self.log_handler.start()
            self.loader.start()

            self.acquiring_signal.emit(True)

        self.quit_lock.release()

        # Start event loop so that signals sent to this thread may be processed
        self.exec()

    def device_disconnected(self):
        self.connected = False
        self.quit()

        # Reset persistent states
        self.pressurize_handler.last_pressurize_bit = None
        self.depressurize_handler.last_depressurize_bit = None
        self.period_handler.last_depressurize_bit = None
        self.period_handler.last_depressurize_event = None
        self.log_handler.last_log_bit = None
        self.sample_sensor_detector.last_result = True
        self.pump_handler.reset()
        self.sentry.reset()

        # Try to reconnect to device
        self.start()

    def quit(self):
        self.connecting = False
        acquired = self.quit_lock.acquire(timeout=10)

        if self.logger is not None:
            self.logger.close()

        self.acquiring_signal.emit(False)
        # Cleanup QThreads
        self.pressurize_handler.quit()
        self.depressurize_handler.quit()
        self.period_handler.quit()
        self.pressure_handler.quit()
        self.pump_handler.quit()
        self.log_handler.quit()
        self.pulse_generator.quit()

        self.pressurize_handler.wait()
        self.depressurize_handler.wait()
        self.period_handler.wait()
        self.pressure_handler.wait()
        self.pump_handler.wait()
        self.log_handler.wait()
        self.pulse_generator.wait()

        # disable pump
        if self.loader.device is not None:
            self.loader.device.stop()

        # equalize pressure
        if self.connected:
            try:
                self.pulse_generator.set_pressurize_low()
            except:
                pass
            try:
                self.pulse_generator.set_depressurize_low()
            except:
                pass
            sleep(1)
            self.pulse_generator.set_pressurize_high()
            self.pulse_generator.set_depressurize_high()

        # close connection
        self.loader.quit()
        self.loader.wait()

        super().quit()
        self.wait()
        if acquired:
            self.quit_lock.release()



from array import array
from threading import Lock
from time import time, sleep
import lzma
import pickle


# Fake device to load example data files
# Used only for testing.
class RawLogReader:
    def __init__(self, filename) -> None:
        self.stop_lock = Lock() # Used to make sure you do not stop the device while reading
        self.sample_rate = 4000
        self.points_to_read = 64
        self.channels_to_read = 8
        self.initial_time = None
        self.current_dio = None
        self.acquiring = None
        self.bytes_to_read = self.channels_to_read * 2 * self.points_to_read
        self.stop()

        # Used to tell how long to wait on reads
        self.read_count = 0
        # File
        self.file = lzma.open(filename, "rb")

    def read_data(self):
        if self.read_count == 0:
            self.initial_time = time()
        else:
            speed = 2
            next_read = self.initial_time + self.read_count * self.points_to_read / (self.sample_rate * speed)
            sleep(max(0, next_read - time()))
        self.read_count += 1

        try:
            # Deserialize each object from the file
            data = pickle.load(self.file)
            # Append the array if it matches the expected structure
            if not isinstance(data, array):
                raise TypeError("Error parsing example data file: Data is not an array.")
            return data
        except EOFError:
            raise RuntimeError("End of file reached.")

    def close_device(self):
        self.file.close()

    # Dummy functions to simulate DI4108USB behavior.
    # None of these do anything more than change member variables.

    def set_dio(self, value=0b1111111, check_echo=True):
        self.current_dio = int(value)

    def start_scan(self):
        # Prevent stopping device while reading
        self.stop_lock.acquire()
        # Start reading
        self.acquiring = True

    def end_scan(self):
        if self.stop_lock.locked():
            self.stop_lock.release()

    def stop(self):
        """
        - stops data acquisition
        - set digital IO to all high
        """
        self.acquiring = False # Signals to stop acquiring
        with self.stop_lock:
            pass
        # Turn all valves off
        self.set_dio(0b1111111, check_echo=False)

    def get_current_dio(self):
        return self.current_dio



from icarus_v2.backend.event_handler import EventHandler
import numpy as np
from icarus_v2.backend.event import Channel, get_channel, Event


# Detects pump strokes and sends to counter
# Take average of average_every points, take the derivative, and smooth it with a gaussian filter.
# Pump strokes are defined as points where this result falls below low_threshold.
# Only the points which initially exceed the thresholds are considered.
class PumpHandler(EventHandler):
    def __init__(self, loader, signal, sample_rate) -> None:
        self.sigma = 5
        self.threshold = 0.01

        self.overlap = 2000 # number of points to add before each chunk
        self.overlap_data = None

        self.event_report_range = (-50, 1500)
        self.event_type = Event.PUMP

        update_rate = 2
        super().__init__(loader, signal, sample_rate, update_rate)


    # Override
    # Loops to transmit data if an event occurs
    def run(self):
        self.running = True
        while self.running:
            # Make sure to fetch a multiple of the average_every points
            data_to_get = int(self.sample_rate / self.update_rate)
            try:
                data, buffer_index = self.reader.read(size=data_to_get, timeout=1)
            except TimeoutError:
                self.running = False
                break

            target_pressure = get_channel(data, Channel.TARGET)

            if self.overlap_data is not None:
                target_pressure = np.concatenate((self.overlap_data, target_pressure))

            x = np.arange(-25, 26)
            y = np.exp(-0.5 * (x / self.sigma) ** 2)
            dy2 = np.roll(y, -1) + np.roll(y, 1) - 2 * y
            dy2 /= dy2.min()
            corr = np.correlate(target_pressure, dy2, mode='same') / np.correlate(target_pressure, y, mode='same')
            stroke = (np.roll(corr, -1) < corr) & (np.roll(corr, 1) < corr) & (corr > self.threshold) & (
                    np.roll(target_pressure, -35) < target_pressure) & (
                             np.roll(target_pressure, 35) < target_pressure) & (target_pressure > 2000)

            # Remove detections near the edges
            if self.overlap_data is not None:
                stroke[:int(self.overlap / 2)] = False
            stroke[-int(self.overlap / 2):] = False

            indices = np.where(stroke)[0]

            # Emit for every dip
            for i in indices:
                event_index = buffer_index + i - self.overlap
                event_data = self.get_event_data(event_index)
                if event_data is not None:
                    sample_rate_kHz = float(self.sample_rate) / 1000
                    chunk_event_index = int( - self.event_report_range[0] * sample_rate_kHz)
                    new_event = Event(self.event_type, event_data, chunk_event_index)
                    self.signal.emit(new_event)

            self.overlap_data = target_pressure[-self.overlap:]

    def reset(self):
        self.overlap_data = None



from PySide6.QtWidgets import QToolBar, QWidget, QSizePolicy, QPushButton
from PySide6.QtGui import QIcon, QAction
from icarus_v2.gui.settings_dialog import SettingsDialog
from PySide6.QtCore import Signal, Qt
from icarus_v2.gui.scrollable_menu import ScrollableMenu


class ToolBar(QToolBar):
    set_mode_signal = Signal(str)

    def __init__(self, parent=None):
        super().__init__(parent=parent)

        self.setMovable(False)
        self.setContextMenuPolicy(Qt.ContextMenuPolicy.PreventContextMenu)

        # Settings button
        self.connected = False
        self.pressure_signal = None
        self.sentry = None
        settings_action = QAction(QIcon(), "Settings", self)
        settings_action.triggered.connect(self.open_settings)
        self.addAction(settings_action)
        self.settings_dialog = None

        # History reset button
        self.history_reset_action = QAction(QIcon(), "Reset History", self)
        self.addAction(self.history_reset_action)

        # Log Loader mode button
        self.change_mode_action = QAction(QIcon(), "Open Logs", self)
        self.change_mode_action.triggered.connect(self.change_log_mode)
        self.addAction(self.change_mode_action)

        # Warning Label
        spacer = QWidget()
        spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        self.addWidget(spacer)

        self.warning_button = QPushButton(self)
        self.warning_button.setFixedWidth(900)
        self.warning_button.setStyleSheet("color: orange; font-size: 12pt; text-align: left;")
        self.addWidget(self.warning_button)

        # Scrollable Menu to show past messages
        self.messages_menu = ScrollableMenu(parent=self.warning_button)
        self.warning_button.setMenu(self.messages_menu)

    def open_settings(self):
        # Open the settings dialog
        self.settings_dialog = SettingsDialog(self.connected, self.pressure_signal, self.sentry)

        def on_dialog_finished():
            self.settings_dialog = None

        self.settings_dialog.finished.connect(on_dialog_finished)

        self.settings_dialog.show()

    def change_log_mode(self):
        if self.change_mode_action.text() == "Open Logs":
            self.change_mode_action.setText("Close Logs")
            self.set_mode_signal.emit("log")

        elif self.change_mode_action.text() == "Close Logs":
            self.change_mode_action.setText("Open Logs")
            self.set_mode_signal.emit("device")

    def set_pressure_signal(self, pressure_signal):
        self.pressure_signal = pressure_signal
        if self.settings_dialog is not None:
            self.settings_dialog.set_pressure_signal(pressure_signal)

    def set_sentry(self, sentry):
        self.sentry = sentry
        if self.settings_dialog is not None:
            self.settings_dialog.set_sentry(sentry)

    def set_connected(self, connected):
        self.connected = connected
        if self.settings_dialog is not None:
            self.settings_dialog.set_connected(connected)

    def display_warning(self, warning, color="orange"):
        self.warning_button.setStyleSheet("color: "+color+"; font-size: 12pt; text-align: left;")
        self.warning_button.setText(warning)
        self.messages_menu.add_message(warning, color)

    '''
    # Clears any log warnings or errors from the toolbar
    # Called whenever the log is closed
    '''
    def clear_log_toolbar(self):
        logs = self.messages_menu.return_matching("LOG:")
        self.messages_menu.remove_items(logs)

        if "LOG:" in self.warning_button.text():
            self.warning_button.setText("")



from PySide6.QtWidgets import QAbstractButton
from PySide6.QtGui import QPixmap, QPainter, QFontMetrics
from PySide6.QtCore import Qt, QTimer


class ImageButton(QAbstractButton):
    def __init__(self, picture, button_text, parent=None, cooldown_ms=0):
        super().__init__(parent)
        self.picture = QPixmap(picture)
        self.button_text = button_text
        self.grey = False
        self.cooldown_duration = cooldown_ms
        self.in_cooldown = False

        # Cooldown timer setup
        self.cooldown_timer = QTimer(self)
        self.cooldown_timer.setSingleShot(True)
        self.cooldown_timer.timeout.connect(self._end_cooldown)

        # Connect signals
        self.pressed.connect(self._grey_out)
        self.released.connect(self._restore)
        self.clicked.connect(self._start_cooldown)

    def _grey_out(self):
        # Temporarily grey out the button
        self.grey = True
        self.update()

    def _restore(self):
        self.grey = False
        self.update()

    def _start_cooldown(self):
        if self.cooldown_duration > 0 and not self.in_cooldown:
            self.in_cooldown = True
            self.cooldown_timer.start(self.cooldown_duration)
            self.update()

    def _end_cooldown(self):
        self.in_cooldown = False
        self.update()

    def mousePressEvent(self, event):
        if self.in_cooldown:
            event.ignore()
        else:
            super().mousePressEvent(event)

    def paintEvent(self, event):
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)

        scaled_pixmap = self.picture.scaled(self.size(), Qt.KeepAspectRatio, Qt.SmoothTransformation)
        x_offset = (self.width() - scaled_pixmap.width()) / 2
        y_offset = (self.height() - scaled_pixmap.height()) / 2
        if self.grey:
            painter.setOpacity(0.6)
        painter.drawPixmap(x_offset, y_offset, scaled_pixmap)
        painter.setOpacity(1)

        # Setting up the font and drawing text
        text = self.button_text
        fm = QFontMetrics(painter.font())
        text_width = fm.boundingRect(text).width()
        text_height = fm.height()
        text_x = (self.width() - text_width) / 2
        text_y = y_offset + (scaled_pixmap.height() / 2) + (text_height / 2)
        painter.drawText(text_x, text_y, text)

        painter.end()



from icarus_v2.backend.event_handler import EventHandler
from icarus_v2.backend.event import Event


# Sends pressure data constantly. Events always occur.
class PressureHandler(EventHandler):
    def __init__(self, loader, signal, sample_rate, update_rate) -> None:
        super().__init__(loader, signal, sample_rate, update_rate)
        self.event_type = Event.PRESSURE


    # Loops to transmit data if an event occurs
    # Overridden because events always occur for this handler and all data is used
    def run(self):
        self.running = True
        while(self.running):
            data_to_get = int(self.sample_rate / self.update_rate)
            try:
                data, buffer_index = self.reader.read(size=data_to_get, timeout=1)
            except TimeoutError:
                self.running = False
                break

            # Transmit data to plot
            new_event = Event(self.event_type, data)
            self.signal.emit(new_event)



import numpy as np
from PySide6.QtCore import QThread, Signal
from icarus_v2.backend.event import get_channel, Channel
from icarus_v2.backend.configuration_manager import ConfigurationManager
from icarus_v2.backend.sentry_error import SentryError
from icarus_v2.backend.sentry_warning import SentryWarning
from time import localtime


# Gets expected values for the experiment checks from the first example_events events of an experiment
class Sentry(QThread):
    warning_signal = Signal(SentryWarning)
    error_signal = Signal(SentryError)

    def __init__(self):
        super().__init__()

        self.config_manager = ConfigurationManager()
        self.settings = self.config_manager.get_settings('sentry_settings')
        self.config_manager.settings_updated.connect(self.update_settings)
        self.current_example_events = None

        self.current_experiment = False
        self.last_error_time = -10
        self.suppress_errors = 2  # suppress errors for 2 seconds after one error

        # values defined after an experiment starts for defaults
        self.example_depress_pressures = []
        self.example_pump_times = [] # last example_events stroke times
        self.recent_pump_times = [] # all times within pump_window
        self.expected_pump_rate = None
        self.expected_pressure_before_depressurize = None
        self.num_pressure_decreases = 0

    # Takes boolean representing the new state of bit 4
    # Resets all expected values
    def handle_experiment(self, event):
        self.current_experiment = not event
        self.expected_pump_rate = None
        self.expected_pressure_before_depressurize = None
        self.example_pump_times = []
        self.recent_pump_times = []
        self.example_depress_pressures = []
        self.num_pressure_decreases = 0
        self.last_error_time = -10
        # update this only here so it is not changed for an ongoing experiment
        self.current_example_events = self.settings['example_events']

    def handle_pump(self, event):
        if self.current_experiment:
            self.example_pump_times.append(event.event_time)
            # Define expected pump rate
            if self.expected_pump_rate is None and len(self.example_pump_times) == self.current_example_events:
                avg = np.diff(self.example_pump_times).mean()
                self.expected_pump_rate = 1 / avg

            # Check for deviation from expected rate (averaged over last example_events)
            if len(self.example_pump_times) > self.current_example_events:
                self.example_pump_times.pop(0)
                pump_rate = 1 / np.diff(self.example_pump_times).mean()
                rate_percent_increase = (pump_rate - self.expected_pump_rate) / self.expected_pump_rate
                if event.event_time - self.last_error_time > self.suppress_errors:
                    if rate_percent_increase > self.settings["max_pump_rate_increase"]:
                        info = {
                            "pump_rate" : pump_rate,
                            "rate_percent_increase" : rate_percent_increase
                        }
                        self.warning_signal.emit(SentryWarning("Pump Rate High",localtime(event.event_time),info))

            # check for max_pumps within pump_window
            self.recent_pump_times.append(event.event_time)
            self.recent_pump_times = [t for t in self.recent_pump_times if t > event.event_time - self.settings["pump_window"]]
            if event.event_time - self.last_error_time > self.suppress_errors:
                if len(self.recent_pump_times) > self.settings["max_pumps_in_window"]:
                    info = {
                        "recent_pump_times" : self.recent_pump_times,
                        "pump_window" : self.settings['pump_window']
                    }
                    self.error_signal.emit(SentryError("Pump Rate High",localtime(event.event_time),info))
                    self.last_error_time = event.event_time

    def handle_depressurize(self, event):
        if self.current_experiment:
            initial_pressure = get_channel(event, Channel.HI_PRE_ORIG)[:event.event_index].mean()

            # define expected pressure before depressurize
            if self.expected_pressure_before_depressurize is None:
                self.example_depress_pressures.append(initial_pressure)
                if len(self.example_depress_pressures) == self.current_example_events:
                    self.expected_pressure_before_depressurize = np.mean(self.example_depress_pressures)

            else:
                percent_change = ((initial_pressure - self.expected_pressure_before_depressurize) /
                                    self.expected_pressure_before_depressurize)
                if percent_change < - self.settings["max_pressure_before_depress_decrease"]:
                    self.num_pressure_decreases += 1
                    if event.event_time - self.last_error_time > self.suppress_errors:
                        if self.num_pressure_decreases >= self.settings["decrease_count_to_error"]:
                            info = {
                                "num_pressure_decreases" : self.num_pressure_decreases
                            }
                            self.error_signal.emit(SentryError("Pressure Decreasing",localtime(event.event_time),info))

                            self.last_error_time = event.event_time
                        else:
                            self.warning_signal.emit(SentryWarning("Pressure Decreasing",localtime(event.event_time)))
                else:
                    self.num_pressure_decreases = 0

                if event.event_time - self.last_error_time > self.suppress_errors:
                    if percent_change > self.settings["max_pressure_before_depress_increase"]:
                        self.warning_signal.emit(SentryWarning("Pressure Increasing",localtime(event.event_time)))

    # On device disconnect
    def reset(self):
        # Exit experiment mode (argument is value of bit 4 and thus is high for normal operation)
        self.handle_experiment(True)

    def update_settings(self, key='sentry_settings'):
        if key == 'sentry_settings':
            self.settings = self.config_manager.get_settings(key)



from __future__ import annotations
import math
from icarus_v2.qdarktheme.rgba import RGBA
from icarus_v2.qdarktheme.hsla import HSLA


class Color:
    """Class handling color code(RGBA and HSLA)."""

    def __init__(self, color_code: RGBA | HSLA) -> None:
        """Initialize color code."""
        self._hsla, self._hsva = None, None
        if isinstance(color_code, RGBA):
            self._rgba = color_code
        elif isinstance(color_code, HSLA):
            self._hsla = color_code
            self._rgba = self._hsla.to_rgba()

    @property
    def rgba(self) -> RGBA:
        """Return rgba."""
        return self._rgba

    @property
    def hsla(self) -> HSLA:
        """Return hsla."""
        return self._hsla if self._hsla else HSLA.from_rgba(self.rgba)

    def __str__(self) -> str:
        """Format Color class.

        e.g. rgba(100, 100, 100, 0.5).
        """
        return str(self.rgba)

    @staticmethod
    def _check_hex_format(hex_format: str) -> None:
        """Check if string is hex format."""
        try:
            hex = hex_format.lstrip("#")
            if not len(hex) in (3, 4, 6, 8):
                raise ValueError
            int(hex, 16)
        except ValueError:
            raise ValueError(
                f'invalid hex color format: "{hex_format}". '
                "Only support following hexadecimal notations: #RGB, #RGBA, #RRGGBB and #RRGGBBAA. "
                "R (red), G (green), B (blue), and A (alpha) are hexadecimal characters "
                "(0-9, a-f or A-F)."
            ) from None

    @staticmethod
    def from_rgba(r: int, g: int, b: int, a: int) -> Color:
        """Convert rgba to Color object."""
        rgba = RGBA(r, g, b, a / 255)
        return Color(rgba)

    @staticmethod
    def from_hex(hex: str) -> Color:
        """Convert hex string to Color object.

        Args:
            color_hex: Color hex string.

        Returns:
            Color: Color object converted from hex.
        """
        Color._check_hex_format(hex)
        hex = hex.lstrip("#")
        r, g, b, a = 255, 0, 0, 1
        if len(hex) == 3:  # #RGB format
            r, g, b = (int(char, 16) for char in hex)
            r, g, b = 16 * r + r, 16 * g + g, 16 * b + b
        if len(hex) == 4:  # #RGBA format
            r, g, b, a = (int(char, 16) for char in hex)
            r, g, b = 16 * r + r, 16 * g + g, 16 * b + b
            a = (16 * a + a) / 255
        if len(hex) == 6:  # #RRGGBB format
            r, g, b = bytes.fromhex(hex)
            a = 1
        elif len(hex) == 8:  # #RRGGBBAA format
            r, g, b, a = bytes.fromhex(hex)
            a = a / 255
        return Color(RGBA(r, g, b, a))

    def to_hex(self) -> str:
        """Convert Color object to hex(#RRGGBBAA).

        Args:
            color: Color object.

        Returns:
            str: Hex converted from Color object.
        """
        r, g, b, a = self.rgba.r, self.rgba.g, self.rgba.b, self.rgba.a
        hex_color = f"{math.floor(r):02x}{math.floor(g):02x}{math.floor(b):02x}"
        if a != 1:
            hex_color += f"{math.floor(a*255):02x}"
        return hex_color

    def to_hex_argb(self) -> str:
        """Convert Color object to hex(#AARRGGBB).

        Args:
            color: Color object.

        Returns:
            str: Hex converted from Color object.
        """
        r, g, b, a = self.rgba.r, self.rgba.g, self.rgba.b, self.rgba.a
        hex_color = "" if a == 1 else f"{math.floor(a*255):02x}"
        hex_color += f"{math.floor(r):02x}{math.floor(g):02x}{math.floor(b):02x}"
        return hex_color

    def to_svg_tiny_color_format(self) -> str:
        """Convert Color object to string for svg.

        QtSvg does not support #RRGGBBAA format.
        Therefore, we need to set the alpha value to `fill-opacity` instead.

        Returns:
            str: RGBA format.
        """
        r, g, b, a = self.rgba
        if a == 1:
            return f'fill="#{self.to_hex()}"'
        return f'fill="rgb({r},{g},{b})" fill-opacity="{a}"'

    def lighten(self, factor: float) -> Color:
        """Lighten color."""
        return Color(HSLA(self.hsla.h, self.hsla.s, self.hsla.l + self.hsla.l * factor, self.hsla.a))

    def darken(self, factor: float) -> Color:
        """Darken color."""
        return Color(HSLA(self.hsla.h, self.hsla.s, self.hsla.l - self.hsla.l * factor, self.hsla.a))

    def transparent(self, factor: float) -> Color:
        """Make color transparent."""
        return Color(RGBA(self.rgba.r, self.rgba.g, self.rgba.b, self.rgba.a * factor))



import lzma
import pickle
from icarus_v2.backend.event import Event
from icarus_v2.backend.sentry_error import SentryError


# Load events from logs
class LogReader:
    def __init__(self, tool_bar=None,) -> None:
        self.events = None
        self.filename = None
        self.logger = None
        self.log_coefficients = None
        self.tool_bar=tool_bar

    def set_logger(self, logger):
        self.logger = logger

    def read_events(self, filename):
        self.log_coefficients = None
        self.events = []
        self.filename = filename

        # If reading current log file, write to disk first
        if self.logger is not None and self.logger.filename is not None and self.filename == self.logger.filename:
            self.logger.flush()

        file = lzma.open(filename, "rb")  # Open file in read binary mode
        try:
            while True:
                # Deserialize event data using pickle
                event_dict = pickle.load(file)
                if "plotting_coefficients" in event_dict.keys():
                    self.log_coefficients = event_dict["plotting_coefficients"]
                elif "error_type" in event_dict.keys():
                    event = event_dict['class_type'](event_dict['error_type'],event_dict['event_time'],event_dict['data'])
                    color="orange"
                    if(type(event) == SentryError):
                        color = "red"

                    self.tool_bar.display_warning("LOG: "+str(event),color)
                else:
                    event = Event(
                        event_dict['event_type'],
                        event_dict['data'],
                        event_dict['event_index'],
                        event_dict['event_time'],
                        event_dict['step_time']
                    )
                    self.events.append(event)
        except EOFError:
            pass  # End of file reached
        file.close()



from __future__ import annotations
import os
import re
from dataclasses import dataclass
from pathlib import Path
from PySide6.QtCore import QStandardPaths
from icarus_v2.qdarktheme.color import Color
from icarus_v2.qdarktheme.svg import Svg

"""Default color values."""
THEME_COLOR_VALUES = {
    "dark": {
        "background": {
            "base": "#2a2a2a",
            "list": {},
            "panel": {"darken": 0.3},
            "popup": {"lighten": 0.3},
            "table": {"darken": 0.5},
            "textarea": {"darken": 0.13},
            "title": {"darken": 0.3},
        },
        "border": {
            "base": "#808080",
            "input": {"transparent": 0},
        },
        "button": {
            "base": "#e4e7eb",
            "button.activeBackground": {"darken": 0.14, "transparent": 0.23},
            "button.hoverBackground": {"darken": 0.1, "transparent": 0.11},
            "defaultButton.activeBackground": {"darken": 0.12},
            "defaultButton.hoverBackground": {"darken": 0.06},
        },
        "foreground": {
            "base": "#e4e7eb",
            "defaultButton.disabledBackground": {"transparent": 0.2},
            "disabled": {"transparent": 0.4},
            "disabledSelectionBackground": {"transparent": 0.2},
            "icon": {"darken": 0.01},
            "icon.unfocused": {"transparent": 0.6},
            "input.placeholder": {"transparent": 0.6},
            "progressBar.disabledBackground": {"transparent": 0.2},
            "slider.disabledBackground": {"transparent": 0.2},
            "sliderTrack.inactiveBackground": {"transparent": 0.1},
        },
        "input.background": "#3f4042",
        "inputButton.hoverBackground": "#ffffff25",
        "linkVisited": "#c58af8",
        "list.alternateBackground": "#ffffff0c",
        "list.hoverBackground": "#ffffff13",
        "menubar.selectionBackground": "#ffffff25",
        "popupItem.checkbox.background": "#ffffff19",
        "popupItem.selectionBackground": "#ffffff22",
        "primary": {
            "base": "#8ab4f7",
            "button.activeBackground": {"darken": 0.14, "transparent": 0.23},
            "button.hoverBackground": {"darken": 0.1, "transparent": 0.11},
            "defaultButton.activeBackground": {"darken": 0.12},
            "defaultButton.hoverBackground": {"darken": 0.06},
            "list.inactiveSelectionBackground": {"lighten": 0.2, "transparent": 0.15},
            "list.selectionBackground": {"darken": 0.2, "transparent": 0.4},
            "progressBar.background": {"darken": 0.1},
            "selection.background": {"darken": 0.2, "lighten": 0.1, "transparent": 0.4},
            "sliderHandle.activeBackground": {"darken": 0.09},
            "table.inactiveSelectionBackground": {"lighten": 0.2, "transparent": 0.18},
            "table.selectionBackground": {"darken": 0.2, "transparent": 0.55},
            "textarea.selectionBackground": {"darken": 0.2, "lighten": 0.1, "transparent": 0.4},
        },
        "scrollbar.background": "#ffffff10",
        "scrollbarSlider.activeBackground": "#ffffff60",
        "scrollbarSlider.background": "#ffffff30",
        "scrollbarSlider.disabledBackground": "#ffffff15",
        "scrollbarSlider.hoverBackground": "#ffffff45",
        "statusBar.background": "#2a2b2e",
        "statusBarItem.activeBackground": "#ffffff34",
        "statusBarItem.hoverBackground": "#ffffff22",
        "tab.activeBackground": "#ffffff00",
        "tab.hoverBackground": "#ffffff18",
        "tabCloseButton.hoverBackground": "#ffffff25",
        "table.alternateBackground": "#ffffff15",
        "tableSectionHeader.background": "#3f4042",
        "textarea.inactiveSelectionBackground": "#ffffff20",
        "toggleButtonOff": {
            "base": "#f45555",
            "button.activeBackground": {"lighten": 0.12},
            "button.hoverBackground": {"lighten": 0.08},
        },
        "toggleButtonOn": {
            "base": "#56bd5d",
            "button.activeBackground": {"lighten": 0.16},
            "button.hoverBackground": {"lighten": 0.08},
        },
        "toolbar.activeBackground": "#ffffff34",
        "toolbar.background": "#2a2a2a",
        "toolbar.hoverBackground": "#ffffff22",
        "tree.inactiveIndentGuidesStroke": "#ffffff35",
        "tree.indentGuidesStroke": "#ffffff60",
        "treeSectionHeader.background": "#3f4042"
    },
    "light": {
        "background": {
            "base": "#f8f9fa",
            "list": {},
            "panel": {"lighten": 0.5},
            "popup": {"lighten": 0.2},
            "table": {"lighten": 0.5},
            "textarea": {"lighten": 0.25},
            "title": {"darken": 0.04},
        },
        "border": {
            "base": "#8E9094",
            "input": {},
        },
        "button": {
            "base": "#4d5157",
            "button.activeBackground": {"darken": 0.14, "transparent": 0.23},
            "button.hoverBackground": {"darken": 0.1, "transparent": 0.11},
            "defaultButton.activeBackground": {"darken": 0.12},
            "defaultButton.hoverBackground": {"darken": 0.06},
        },
        "foreground": {
            "base": "#1A1E24",
            "defaultButton.disabledBackground": {"transparent": 0.25},
            "disabled": {"transparent": 0.4},
            "disabledSelectionBackground": {"transparent": 0.25},
            "icon": {"darken": 0.05},
            "icon.unfocused": {"transparent": 0.6},
            "input.placeholder": {"transparent": 0.6},
            "progressBar.disabledBackground": {"transparent": 0.25},
            "slider.disabledBackground": {"transparent": 0.25},
            "sliderTrack.inactiveBackground": {"transparent": 0.2},
        },
        "input.background": "#f8f9fa",
        "inputButton.hoverBackground": "#00000018",
        "linkVisited": "#660098",
        "list.alternateBackground": "#00000009",
        "list.hoverBackground": "#00000013",
        "menubar.selectionBackground": "#00000020",
        "popupItem.checkbox.background": "#00000019",
        "popupItem.selectionBackground": "#00000022",
        "primary": {
            "base": "#1a73e8",
            "button.activeBackground": {"darken": 0.03, "transparent": 0.24},
            "button.hoverBackground": {"darken": 0.03, "transparent": 0.1},
            "defaultButton.activeBackground": {"lighten": 0.3},
            "defaultButton.hoverBackground": {"lighten": 0.1},
            "list.inactiveSelectionBackground": {"darken": 0.43, "transparent": 0.09},
            "list.selectionBackground": {"lighten": 0.2, "transparent": 0.35},
            "progressBar.background": {"lighten": 0.2},
            "selection.background": {"lighten": 0.3, "transparent": 0.5},
            "sliderHandle.activeBackground": {"lighten": 0.2},
            "table.inactiveSelectionBackground": {"darken": 0.43, "transparent": 0.09},
            "table.selectionBackground": {"lighten": 0.1, "transparent": 0.5},
            "textarea.selectionBackground": {"lighten": 0.3, "transparent": 0.5},
        },
        "scrollbar.background": "#00000010",
        "scrollbarSlider.activeBackground": "#00000060",
        "scrollbarSlider.background": "#00000040",
        "scrollbarSlider.disabledBackground": "#00000015",
        "scrollbarSlider.hoverBackground": "#00000050",
        "statusBar.background": "#dfe1e5",
        "statusBarItem.activeBackground": "#00000024",
        "statusBarItem.hoverBackground": "#00000015",
        "tab.activeBackground": "#00000000",
        "tab.hoverBackground": "#00000015",
        "tabCloseButton.hoverBackground": "#00000020",
        "table.alternateBackground": "#00000012",
        "tableSectionHeader.background": "#dadce0",
        "textarea.inactiveSelectionBackground": "#00000015",
        "toggleButtonOff": {
            "base": "#f45555",
            "button.activeBackground": {"lighten": 0.12},
            "button.hoverBackground": {"lighten": 0.08},
        },
        "toggleButtonOn": {
            "base": "#56bd5d",
            "button.activeBackground": {"lighten": 0.16},
            "button.hoverBackground": {"lighten": 0.08},
        },
        "toolbar.activeBackground": "#00000024",
        "toolbar.background": "#ebebeb",
        "toolbar.hoverBackground": "#00000015",
        "tree.inactiveIndentGuidesStroke": "#00000030",
        "tree.indentGuidesStroke": "#00000050",
        "treeSectionHeader.background": "#dadce0",
    }
}
ACCENT_COLORS = {
    "dark": {
        "light_green": '#45BF55',
        "magenta": '#AB47BC',
        "blue": '#004B8D',
        "yellow": '#FFDC00',
        "cyan": '#59D8E6',
        "red": '#B9121B',

        # "blue": "#004B8D",
        "graphite": "#898a8f",
        "green": "#4caf50",
        "orange": "#ff9800",
        "pink": "#c7457f",
        "purple": "#af52bf",
        # "red": "#f6685e",
        # "yellow": "#ffeb3b",
    },
    "light": {
        "light_green": '#4caf50',
        "magenta": '#AB47BC',
        "blue": '#004B8D',
        "yellow": '#ff9800',
        "cyan": '#004B8D',
        "red": '#f44336',

        # "blue": "#1a73e8",
        "graphite": "#898a8f",
        "green": "#4caf50",
        "orange": "#ff9800",
        "pink": "#c7457f",
        "purple": "#9c27b0",
        # "red": "#f44336",
        # "yellow": "#f4c65f",
    },
}


def color(color_info: str | dict[str, str | dict], state: str | None = None) -> Color:
    """Filter for template engine. This filter convert color info data to color object."""
    if isinstance(color_info, str):
        return Color.from_hex(color_info)

    base_color_format: str = color_info["base"]  # type: ignore
    color = Color.from_hex(base_color_format)

    if state is None:
        return color

    transforms = color_info[state]
    return Color.from_hex(transforms) if isinstance(transforms, str) else _transform(color, transforms)


def url(color: Color, id: str, rotate: int = 0) -> str:
    """Filter for template engine. This filter create url for svg and output svg file."""
    cache_dir = QStandardPaths.writableLocation(QStandardPaths.CacheLocation)
    icons_dir = Path(cache_dir) / "icons"
    svg_path = icons_dir / f"{id}_{color.to_hex()}_{rotate}.svg"
    url = f"url({svg_path.as_posix()})"

    # Create the icons directory if it doesn't exist
    icons_dir.mkdir(parents=True, exist_ok=True)

    if svg_path.exists():
        return url
    svg = Svg(id).colored(color).rotate(rotate)
    svg_path.write_text(str(svg))
    return url


def env(
    text, value: str, version: str | None = None, qt: str | None = None, os: str | None = None
) -> str:
    return value.replace("${}", str(text))


def corner(corner_shape: str, size: str) -> str:
    """Filter for template engine. This filter manage corner shape."""
    return size if corner_shape == "rounded" else "0"


def _transform(color: Color, color_state: dict[str, float]) -> Color:
    if color_state.get("transparent"):
        color = color.transparent(color_state["transparent"])
    if color_state.get("darken"):
        color = color.darken(color_state["darken"])
    if color_state.get("lighten"):
        return color.lighten(color_state["lighten"])
    return color


@dataclass(unsafe_hash=True, frozen=True)
class _Placeholder:
    match_text: str
    value: str | int | float
    filters: tuple[str]


class Template:
    """Class that handles template text like jinja2."""

    _PLACEHOLDER_RE = re.compile(r"{{.*?}}")
    _STRING_RE = re.compile(r"""(('([^'\\]*(?:\\.[^'\\]*)*)'|"([^"\\]*(?:\\.[^"\\]*)*)"){1,3})""", re.S)

    def __init__(self, text: str, filters: dict):
        """Initialize Template class."""
        self._target_text = text
        self._filters = filters

    @staticmethod
    def _to_py_value(text: str):
        try:
            return int(text)
        except ValueError:
            try:
                return float(text)
            except ValueError:
                return text

    @staticmethod
    def _parse_placeholders(text: str):
        placeholders: set[_Placeholder] = set()
        for match in re.finditer(Template._PLACEHOLDER_RE, text):
            match_text = match.group()
            contents, *filters = match_text.strip("{}").replace(" ", "").split("|")
            value = Template._to_py_value(contents)
            placeholders.add(_Placeholder(match_text, value, tuple(filters)))
        return placeholders

    def _run_filter(self, value: str | int | float, filter_text: str):
        contents = filter_text.split("(")
        if len(contents) == 1:
            return self._filters[contents[0]](value)

        filter_name, arg_text = contents[0], contents[1].rstrip(")")

        # Split arguments by commas, then key-value pairs by '='
        arguments = {}
        if arg_text:
            for arg in arg_text.split(","):
                arg = arg.strip()
                if "=" in arg:
                    # Ensure there is exactly one '=' and split by it
                    key_val_pair = arg.split("=", 1)
                    key = key_val_pair[0].strip()
                    val = key_val_pair[1].strip().strip('"').strip("'")  # Handle string values
                    arguments[key] = val
                else:
                    # Handle positional arguments if needed
                    arguments[arg] = True

        return self._filters[filter_name](value, **arguments)

    def render(self, replacements: dict) -> str:
        """Render replacements."""
        placeholders = Template._parse_placeholders(self._target_text)
        new_replacements: dict[str, str] = {}
        for placeholder in placeholders:
            value = placeholder.value
            if type(value) is str and len(value) != 0:
                value = replacements.get(value)
            if value is None:
                raise AssertionError(
                    f"There is no replacements for: {placeholder.value} in {placeholder.match_text}"
                )
            for filter in placeholder.filters:
                value = self._run_filter(value, filter)
            new_replacements[placeholder.match_text] = str(value)
        return multi_replace(self._target_text, new_replacements)


def multi_replace(target: str, replacements: dict[str, str]) -> str:
    """Given a string and a replacement map, it returns the replaced string.

    See https://gist.github.com/bgusach/a967e0587d6e01e889fd1d776c5f3729.

    Args:
        target: String to execute replacements on.
        replacements: Replacement dictionary {value to find: value to replace}.

    Returns:
        str: Target string that replaced with `replacements`.
    """
    if len(replacements) == 0:
        return target

    replacements_sorted = sorted(replacements, key=len, reverse=True)
    replacements_escaped = [re.escape(i) for i in replacements_sorted]
    pattern = re.compile("|".join(replacements_escaped))
    return pattern.sub(lambda match: replacements[match.group()], target)


def _color_values(theme: str) -> dict[str, str | dict]:
    try:
        return THEME_COLOR_VALUES[theme]
    except KeyError:
        raise ValueError(f'invalid argument, not a dark, light or auto: "{theme}"') from None


def load_stylesheet(theme: str = "dark", corner_shape: str = "rounded",) -> str:
    color_values = _color_values(theme)
    stylesheet = load_base_stylesheet()

    # Build stylesheet
    template = Template(
        stylesheet,
        {"color": color, "corner": corner, "env": env, "url": url},
    )
    replacements = dict(color_values, **{"corner-shape": corner_shape})
    return template.render(replacements)


def load_base_stylesheet():
    current_dir = os.path.dirname(os.path.abspath(__file__))
    file_path = os.path.join(current_dir, 'styles.qss')
    with open(file_path, 'r') as f:
        return f.read()



import usb.core
from array import array
from threading import Lock


# Device information
DI_4108_VENDOR_ID = 0x0683
DI_4108_PRODUCT_ID = 0x4108


# Interface to the Di4108 USB device, capable of reading from its DIO and Analog channels and sending instructions.
class DataqInterface:
    def __init__(self) -> None:
        """
        Initializes a DataqInterface object and sets up device for reading.
        """
        self.stop_lock = Lock()  # Used to make sure you do not stop the device while reading

        self.device = None
        self.endpoint_out = None
        self.endpoint_in = None
        self.usb_buff = None
        self.sample_rate = None
        self.points_to_read = None
        self.channels_to_read = None
        self.bytes_to_read = None
        self.current_dio = None
        self.acquiring = None

        self.find_device()
        self.setup_device()

    def find_device(self):
        """
        Finds the usb device.

        :return: True if connection is successful, False otherwise.
        """
        try:
            # Find the USB device
            self.device = usb.core.find(idVendor=DI_4108_VENDOR_ID, idProduct=DI_4108_PRODUCT_ID)
            if self.device is None:
                raise RuntimeError("USB device not found. Please ensure the device is connected.")
            else:
                # Attempt to read from the device as a simple permission test
                cfg = self.device.get_active_configuration()
                intf = cfg[(0, 0)]
                ep = usb.util.find_descriptor(
                    intf,
                    custom_match=lambda err: usb.util.endpoint_direction(err.bEndpointAddress) == usb.util.ENDPOINT_IN
                )                
                assert ep is not None
                # If no exception was raised, assume permissions are adequate
                return True
        except usb.core.USBError as e:
            if e.errno == 13:
                raise RuntimeError("Insufficient permissions to access the USB device. " +
                                   "Please set up udev rules by running setup/install_udev_rules.sh.")
            else:
                raise RuntimeError(e)

    def setup_device(self):
        # Reinitialize device
        self.device.reset()
        # Set the first configuration as active
        self.device.set_configuration()

        # Get and setup endpoints
        cfg = self.device.get_active_configuration()
        intf = cfg[(0, 0)]
        def match_out(e): return usb.util.endpoint_direction(e.bEndpointAddress) == usb.util.ENDPOINT_OUT
        def match_in(e): return usb.util.endpoint_direction(e.bEndpointAddress) == usb.util.ENDPOINT_IN
        self.endpoint_out = usb.util.find_descriptor(intf, custom_match=match_out)
        self.endpoint_in = usb.util.find_descriptor(intf, custom_match=match_in)
        self.endpoint_out.wMaxPacketSize = 7200000
        self.endpoint_in.wMaxPacketSize = 7200000
        self.endpoint_in.bmAttributes = 1
        self.endpoint_in.bInterval = 100
        self.usb_buff = int(self.endpoint_in.wMaxPacketSize/100)

        # Stop in case device was left running
        self.stop()
        # Set all dio ports to switches
        self.send_cmd("endo 127")
        # Define binary output mode
        self.send_cmd("encode 0")
        # Set packet size = 1024 bytes
        self.send_cmd("ps 6")
        # Set all channels to CIC filtering
        self.send_cmd("filter * 1")

        # Set slist which is order of channels to sample. See device protocol for more info.
        # This is A0-A6 at +-10V, Digital
        slist = ["0 0", "1 1", "2 2", "3 3", "4 4", "5 5", "6 6", "7 8"]
        for i in slist:
            self.send_cmd("slist " + i)

        # Sample rate (Hz) = 60,000,000 / (srate * dec * deca)
        # Device reports 1 value per (dec * deca) readings. (default is by CIC filtering)
        srate = 3000
        dec = 5
        self.sample_rate = 60000000 / (srate * dec)
        self.send_cmd('srate ' + str(srate))
        self.send_cmd('dec ' + str(dec))

        # Calculate number of bytes to read
        self.points_to_read = 64
        self.channels_to_read = len(slist)
        # 2 bytes per channel
        self.bytes_to_read = self.channels_to_read * 2 * self.points_to_read

    def send_cmd(self, command, check_echo=True):
        """
        Sends a command to the USB device.

        :param command: The command to be sent.
        :param check_echo: Whether to compare the return of the device. Do not use while acquiring data.
        """
        self.device.write(self.endpoint_out, (command+'\r').encode('utf-8'))

        # Expect a response unless the device is currently reading
        # Not reading a response can leave the response in the buffer which can cause data processing issues or overflow
        if not self.acquiring:
            response = self.read()
            if response is not None:
                response = bytes(response).decode('utf-8', errors='ignore').strip('\0')
            expected_response = command+'\r'
            if check_echo and response != expected_response:
                print(f"Error sending command: Response \"{expected_response.strip()}\"" +
                      f"expected but \"{response.strip()}\" received.")

    # Default value - pump off, valves closed, not logging
    def set_dio(self, value=0b1111111, check_echo=True):
        """
        Sets digital input/output state.

        :param value: States to set.
        :param check_echo: Whether to compare the return of the device. Do not use while acquiring data.
        """
        self.current_dio = int(value)
        self.send_cmd("dout " + str(int(value)), check_echo=check_echo)

    def read(self, size=None, timeout=2000):
        if size is None:
            size = self.usb_buff
        data = self.device.read(self.endpoint_in, size, timeout=timeout)
        return data

    def start_scan(self):
        # Prevent stopping device while reading
        self.stop_lock.acquire()
        # Start reading
        self.acquiring = True

        self.send_cmd('start', check_echo=False)

    def end_scan(self):
        if self.stop_lock.locked():
            self.stop_lock.release()

    def read_data(self):
        data = self.read(self.bytes_to_read)
        if data is None:
            return None

        # Check for buffer overflow
        # Represents a received value of "stop 01".
        stop_01_array = array('B', [115, 116, 111, 112, 32, 48, 49, 13, 0])
        if stop_01_array == data[-9:]:
            raise RuntimeError("Error: Buffer overflow on physical device. Scanning Stopped.")

        return data

    def close_device(self):
        """
        Closes the USB device.
        """
        if self.device is not None:
            usb.util.dispose_resources(self.device)

    def stop(self):
        """
        - stops data acquisiion
        - set digital IO to all high
        """
        self.acquiring = False # Signals to stop acquiring
        with self.stop_lock:
            self.send_cmd("stop", check_echo=False)
        # Turn all valves off
        self.set_dio(0b1111111, check_echo=False)

    def get_current_dio(self):
        return self.current_dio



from time import strftime

class SentryError(RuntimeError):
    def __init__(self, error_type, time, info=None):
        self.error_type = error_type
        self.time = time
        self.info = info

    def __str__(self):
        if self.error_type == "Pump Rate High":
            return (
                f"Error: {len(self.info['recent_pump_times'])} pump strokes occurred within {self.info['pump_window']} seconds.")
        elif self.error_type == "Pressure Decreasing":
            return (
                f"Error: pressure decreased {self.info['num_pressure_decreases']} times in a row at "
                f"{strftime('%H:%M:%S', self.time)}. Likely leak")
        
        return "Unknown Error"


from icarus_v2.backend.event_handler import EventHandler
import numpy as np
from icarus_v2.backend.event import Channel, get_channel


# Detects log bit change event and transmits to logger
class LogHandler(EventHandler):
    def __init__(self, loader, signal, sample_rate, update_rate) -> None:
        super().__init__(loader, signal, sample_rate, update_rate)
        self.last_log_bit = None # variable to keep track of edges of data chunks in case an event lines up with the start of a chunk


    # Overridden because this should be able to handle multiple events in one chunk whereas the base class cannot
    def run(self):
        self.running = True
        while(self.running):
            data_to_get = int(self.sample_rate / self.update_rate)
            try:
                data, buffer_index = self.reader.read(size=data_to_get, timeout=1)
            except TimeoutError:
                self.running = False
                self.last_log_bit = None
                break

            # use XOR with offset array to find indeces where there is a state change
            log_data = get_channel(data, Channel.LOG)

            # Start of data stream always starts a log file
            if self.last_log_bit is None:
                self.signal.emit(log_data[0])
                self.last_log_bit = log_data[0]

            log_offset = np.insert(log_data, 0, self.last_log_bit)[:-1]

            # Find indeces where there is a state change
            # np.where returns a tuple, so take the first element
            changes = np.where(log_offset ^ log_data)[0]

            for index in changes:
                self.signal.emit(log_data[index])

            self.last_log_bit = log_data[-1]


    # Setting last_log_bit ensures that a new log file is started when device is reconnected
    def quit(self):
        self.running = False
        self.last_log_bit = None



from time import strftime

class SentryWarning(Warning):
    def __init__(self, error_type, time, info=None):
        self.error_type = error_type
        self.time = time
        self.info = info

    def __str__(self):
        if self.error_type == "Pump Rate High":
            return (
                f"Warning: pump rate at {strftime('%H:%M:%S', self.time)} is "
                f"{self.info['pump_rate']:.2f}, which is {self.info['rate_percent_increase']:.2f}% over the expected rate from the "
                "beginning of the experiment. Possible leak.")
        elif self.error_type == "Pressure Decreasing":
            return (
                f"Warning: pressure decreasing at "
                f"{strftime('%H:%M:%S', self.time)}. Possible leak.")
        elif self.error_type == "Pressure Increasing":
            return (
                f"Warning: pressure increasing at "
                f"{strftime('%H:%M:%S', self.time)}. Possible leak.")
        
        return "Unknown Warning"


from icarus_v2.backend.event_handler import EventHandler
import numpy as np
from icarus_v2.backend.event import Event, Channel, get_channel


# Detects a depressurize event and transmits data to plot
class DepressurizeHandler(EventHandler):
    def __init__(self, loader, signal, sample_rate, update_rate, event_report_range) -> None:
        super().__init__(loader, signal, sample_rate, update_rate)
        self.event_report_range = event_report_range # tuple of range of ms around an event to report e.g. (-10,140)
        self.last_depressurize_bit = None # variable to keep track of edges of data chunks in case an event lines up with the start of a chunk
        self.event_type = Event.DEPRESSURIZE


    # Data: one chunk from the reader
    # Returns whether an event occurs and the index of the event
    def detect_event(self, data):
        depre_valve = get_channel(data, Channel.DEPRE_VALVE)
        # valve array with prior value inserted and last value popped
        valve_offset = np.insert(depre_valve, 0, self.last_depressurize_bit)[:-1]

        # Find indeces where there is a low bit preceeded by a high bit
        # np.where returns a tuple, so take the first element
        low_indeces = np.where(valve_offset & ~depre_valve)[0]

        # All values are high. No event.
        if len(low_indeces) == 0:
            event = False
            index = -1

        # Event occured
        else:
            event = True
            index = low_indeces[0]

            # Raise warning if 2 low pulses detected in same chunk. Under default settings this will never happen unless 2 pulses are made within 33ms of each other
            if len(low_indeces) > 1:
                raise RuntimeError("Depressurize event dropped. Two events occured in same data chunk.")

        self.last_depressurize_bit = depre_valve[-1]
        return event, index


    # Returns data to graph
    def handle_event(self, event_index):
        data = self.get_event_data(event_index)

        before, after = self.event_report_range
        sample_rate_kHz = float(self.sample_rate) / 1000
        event_index = int( - before * sample_rate_kHz)

        return data, event_index



from __future__ import annotations
import colorsys
from icarus_v2.qdarktheme.rgba import RGBA


def _round_float(number: float, decimal_points: int = 3) -> float:
    decimal = 10**decimal_points
    return round(number * decimal) / decimal


class HSLA:
    def __init__(self, h: int, s: float, l: float, a: float = 1) -> None:  # noqa: E741
        self._h = max(min(360, h), 0) | 0
        self._s = _round_float(max(min(1, s), 0))
        self._l = _round_float(max(min(1, l), 0))
        self._a = _round_float(max(min(1, a), 0))

    def __eq__(self, other: HSLA) -> bool:
        """Returns true if `h`, `s`, `l` and `a` are all the same."""
        return [self.h, self.s, self.l, self.a] == [other.h, other.s, other.l, other.a]

    @property
    def h(self) -> int:
        return self._h

    @property
    def s(self) -> float:
        return self._s

    @property
    def l(self) -> float:  # noqa: E741, E743
        return self._l

    @property
    def a(self) -> float:
        return self._a

    @staticmethod
    def from_rgba(rgba: RGBA) -> HSLA:
        hls = colorsys.rgb_to_hls(rgba.r / 255, rgba.g / 255, rgba.b / 255)
        return HSLA(int(hls[0] * 360), hls[2], hls[1], rgba.a)

    def to_rgba(self) -> RGBA:
        rgb = colorsys.hls_to_rgb(self.h / 360, self.l, self.s)
        return RGBA(round(rgb[0] * 255), round(rgb[1] * 255), round(rgb[2] * 255), self.a)



import csv
import numpy as np
import pyqtgraph as pg
from PySide6.QtCore import QCoreApplication
import pyqtgraph.exporters


translate = QCoreApplication.translate


# Custom override of CSVExporter class
# Adds functionality to override auto-generated csv headers with custom ones
class CSVExporter(pg.exporters.CSVExporter):
    def __init__(self, item):
        pg.exporters.CSVExporter.__init__(self, item)

    def export(self, filename=None):
        # Initialize variables
        header = ["Time"]  # Start with Time column
        time_columns = []
        data_columns = []
        data_names = []

        # Collect data from all plot items
        for item in self.item.items:
            if hasattr(item, 'implements') and item.implements('plotData'):
                cd = item.getOriginalDataset()
                time_columns.append(cd[0])  # Store x-axis data
                data_columns.append(cd[1])  # Store y-axis data
                data_names.append(item.name().replace(' ', '_'))
                header.append(item.name().replace(' ', '_'))

        # Get unique sorted timestamps from all time columns
        filtered_columns = [arr for arr in time_columns if arr is not None]
        merged_time = np.unique(np.concatenate(filtered_columns)) if filtered_columns else np.array([])

        if len(merged_time) > 0:
            merged_time.sort()

            # Create a lookup dictionary for each data series
            series_data = {name: {} for name in data_names}
            for i in range(len(time_columns)):
                if time_columns[i] is not None:
                    series_data[data_names[i]] = dict(zip(time_columns[i], data_columns[i]))

            # First add the time column
            row_data = [merged_time]

            # Add data columns
            for name in data_names:
                column_data = []
                for t in merged_time:
                    value = series_data[name].get(t, "")
                    # Convert DIO values to int if necessary
                    if "DIO" in name and value != "":
                        value = int(bool(value))
                    column_data.append(value)
                row_data.append(column_data)

            # Transpose the data for writing
            output_data = list(map(list, zip(*row_data)))

        else:
            output_data = None

        # Write to CSV
        with open(filename, 'w', newline='') as csvfile:
            writer = csv.writer(csvfile, quoting=csv.QUOTE_MINIMAL)
            writer.writerow(header)

            if output_data is not None:
                writer.writerows(output_data)

        self.data.clear()



import json
import os
import importlib
import threading
from copy import deepcopy
from PySide6.QtCore import Signal, QObject, QStandardPaths, QMetaObject, Q_ARG, Qt
from icarus_v2.backend.event import Channel


# Responsible for loading and saving settings
# Thread-safe singleton
class ConfigurationManager(QObject):
    _instance = None
    _lock = threading.Lock()  # Lock for thread safety
    FILENAME = "settings.json"
    # Signal to let subscribers know that settings[key] was changed
    settings_updated = Signal(str)

    def __new__(cls):
        with cls._lock:
            if cls._instance is None:
                cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self):
        if not hasattr(self, "initialized"):
            super().__init__()
            self._initialize()

    def _initialize(self):
        self.initialized = True

        # Get the application configuration path
        self.config_path = QStandardPaths.writableLocation(QStandardPaths.AppConfigLocation)
        os.makedirs(self.config_path, exist_ok=True)  # Create the directory if it doesn't exist

        self.filename = os.path.join(self.config_path, self.FILENAME)

        # Load application settings
        if not os.path.isfile(self.filename):
            self.settings = self.get_default_settings()
            self.save_settings()  # Save default settings to the config file
        else:
            with open(self.filename, "r") as file:
                # Dictionary
                self.settings = json.load(file)

    def get_settings(self, key):
        with self._lock:
            # Get default value if not defined
            if key not in self.settings:
                default = self.get_default_settings()
                self.settings[key] = default[key]

            value = deepcopy(self.settings[key])

            # Convert to enum rather than int
            if key == 'plotting_coefficients':
                value = {Channel(int(k)): v for k, v in value.items()}

        return value

    def save_settings(self, key=None, value=None, emit=True):
        with self._lock:
            if key is not None and value is not None:
                # Convert to int rather than enum
                if key == 'plotting_coefficients':
                    value = {k.value: v for k, v in value.items()}
                self.settings[key] = value

            with open(self.filename, "w") as file:
                json.dump(self.settings, file, indent=4)

        if emit and key is not None:
            # Emit in a thread-safe way
            QMetaObject.invokeMethod(self, "settings_updated", Qt.QueuedConnection, Q_ARG(str, key))

    @staticmethod
    def get_default_settings():
        # Load default settings from icarus_v2.resources
        with importlib.resources.path('icarus_v2.resources', 'default_settings.json') as path:
            with open(path, 'r') as file:
                settings = json.load(file)
        return settings



from PySide6.QtWidgets import QWidget, QVBoxLayout
import pyqtgraph as pg
from icarus_v2.backend.event import Event, Channel, HistStat
import numpy as np
from icarus_v2.gui.styled_plot_widget import StyledPlotWidget
from icarus_v2.backend.configuration_manager import ConfigurationManager


MAX_PRESSURE_INTERVAL = 60


class HistoryPlot(QWidget):
    def __init__(self):
        super().__init__()

        self.config_manager = ConfigurationManager()
        self.coefficients = None
        self.log_coefficients = None
        self.update_settings("plotting_coefficients")
        self.config_manager.settings_updated.connect(self.update_settings)
        self.initial_time = None
        self.can_plot_pressurize = None
        self.last_pressure_time = None
        self.limits = None

        # Pressure plot
        self.pressure_plot = StyledPlotWidget(x_zoom=True)
        self.pressure_plot.set_title("Pressure")
        self.pressure_plot.set_y_label('Pressure (kBar)')
        self.pressure_plot.set_x_label('Time (s)')
        self.pressure_plot.setYRange(0, 3)
        self.pressure_plot.getPlotItem().getAxis("left").setWidth(45)
        self.pressure_plot.add_line(HistStat.O_PRESS)
        self.pressure_plot.add_line(HistStat.S_PRESS)
        self.pressure_plot.add_statistic(HistStat.O_PRESS, lambda x: x[-1], "Last: {:.3f}")
        self.pressure_plot.add_statistic(HistStat.O_PRESS, np.mean, "Avg: {:.3f}")

        # Slope plot
        self.slope_plot = StyledPlotWidget(x_zoom=True)
        self.slope_plot.set_title("Pressure Change Slope")
        self.slope_plot.set_y_label('Slope (kBar/ms)')
        self.slope_plot.set_x_label('Time (s)')
        self.slope_plot.setYRange(-1.1, 1.1)
        self.slope_plot.getPlotItem().getAxis("left").setWidth(45)
        self.slope_plot.add_line(HistStat.DO_SLOPE)
        self.slope_plot.add_line(HistStat.DS_SLOPE)
        self.slope_plot.add_line(HistStat.PO_SLOPE)
        self.slope_plot.add_line(HistStat.PS_SLOPE)
        # Connect the x-axis of all plots for zooming and panning
        self.slope_plot.setXLink(self.pressure_plot)
        self.slope_plot.add_statistic(HistStat.PO_SLOPE, lambda x: x[-1], "Last: {:.2f}")
        self.slope_plot.add_statistic(HistStat.PO_SLOPE, np.mean, "Avg: {:.2f}")
        self.slope_plot.add_statistic(HistStat.DO_SLOPE, lambda x: x[-1], "Last: {:.2f}")
        self.slope_plot.add_statistic(HistStat.DO_SLOPE, np.mean, "Avg: {:.2f}")

        # Switch time plot
        self.switch_time_plot = StyledPlotWidget(x_zoom=True)
        self.switch_time_plot.set_title("Switch Time")
        self.switch_time_plot.set_y_label('Time (ms)')
        self.switch_time_plot.set_x_label('Time (s)')
        self.switch_time_plot.setYRange(0, 45)
        self.switch_time_plot.getPlotItem().getAxis("left").setWidth(45)
        self.switch_time_plot.add_line(HistStat.DO_SWITCH)
        self.switch_time_plot.add_line(HistStat.DS_SWITCH)
        self.switch_time_plot.add_line(HistStat.PO_SWITCH)
        self.switch_time_plot.add_line(HistStat.PS_SWITCH)
        # Connect the x-axis of all plots for zooming and panning
        self.switch_time_plot.setXLink(self.pressure_plot)
        self.switch_time_plot.add_statistic(HistStat.PO_SWITCH, lambda x: x[-1], "Last: {:.2f}")
        self.switch_time_plot.add_statistic(HistStat.PO_SWITCH, np.mean, "Avg: {:.2f}")
        self.switch_time_plot.add_statistic(HistStat.DO_SWITCH, lambda x: x[-1], "Last: {:.2f}")
        self.switch_time_plot.add_statistic(HistStat.DO_SWITCH, np.mean, "Avg: {:.2f}")

        # Lines for displaying log view time
        self.press_time_press = None
        self.press_time_slope = None
        self.press_time_switch = None
        self.depress_time_press = None
        self.depress_time_slope = None
        self.depress_time_switch = None

        # Set layout
        self.layout = QVBoxLayout()
        self.layout.setSpacing(0)
        self.layout.addWidget(self.pressure_plot)
        self.layout.addWidget(self.slope_plot)
        self.layout.addWidget(self.switch_time_plot)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.layout)

        self.reset_history()

    def reset_history(self):
        self.initial_time = None
        self.can_plot_pressurize = True
        self.last_pressure_time = None

        self.pressure_plot.reset()
        self.slope_plot.reset()
        self.switch_time_plot.reset()

        self.limits = (0, 10)
        self.reset_limits()

        # Lines for displaying log view time
        self.pressure_plot.removeItem(self.press_time_press)
        self.slope_plot.removeItem(self.press_time_slope)
        self.switch_time_plot.removeItem(self.press_time_switch)
        self.pressure_plot.removeItem(self.depress_time_press)
        self.slope_plot.removeItem(self.depress_time_slope)
        self.switch_time_plot.removeItem(self.depress_time_switch)

        self.pressure_plot.setXRange(self.limits[0], self.limits[1])

    def add_event(self, event):
        if event.event_type == Event.PRESSURE:
            if self.last_pressure_time is None or event.event_time - self.last_pressure_time <= MAX_PRESSURE_INTERVAL:
                return

        # Define start time of plots
        if self.initial_time is None:
            self.initial_time = event.event_time

        # plot only one pressurize per pressurize. plots are noisy with multiple pressurizes.
        if event.event_type == Event.PRESSURIZE:
            if not self.can_plot_pressurize:
                return
            self.can_plot_pressurize = False
        elif event.event_type == Event.DEPRESSURIZE:
            self.can_plot_pressurize = True

        # Check if currently fully zoomed out
        currently_zoomed_out = self.pressure_plot.get_view_state()

        # use log coefficients if defined
        coefficients = self.coefficients if self.log_coefficients is None else self.log_coefficients

        # Time since first event
        time = event.event_time - self.initial_time

        if event.event_type == Event.PRESSURIZE:
            self.slope_plot.append_points({
                HistStat.PO_SLOPE: event.get_event_info(HistStat.PO_SLOPE) * coefficients[HistStat.PO_SLOPE],
                HistStat.PS_SLOPE: event.get_event_info(HistStat.PS_SLOPE) * coefficients[HistStat.PS_SLOPE]
            }, time)

            self.switch_time_plot.append_points({
                HistStat.PO_SWITCH: event.get_event_info(HistStat.PO_SWITCH) * coefficients[HistStat.PO_SWITCH],
                HistStat.PS_SWITCH: event.get_event_info(HistStat.PS_SWITCH) * coefficients[HistStat.PS_SWITCH]
            }, time)

        if event.event_type == Event.DEPRESSURIZE or event.event_type == Event.PRESSURE:
            self.pressure_plot.append_points({
                HistStat.O_PRESS: event.get_event_info(HistStat.O_PRESS) * coefficients[HistStat.O_PRESS],
                HistStat.S_PRESS: event.get_event_info(HistStat.S_PRESS) * coefficients[HistStat.S_PRESS]
            }, time)

            self.last_pressure_time = event.event_time

        if event.event_type == Event.DEPRESSURIZE:
            self.slope_plot.append_points({
                HistStat.DO_SLOPE: event.get_event_info(HistStat.DO_SLOPE) * coefficients[HistStat.DO_SLOPE],
                HistStat.DS_SLOPE: event.get_event_info(HistStat.DS_SLOPE) * coefficients[HistStat.DS_SLOPE]
            }, time)

            self.switch_time_plot.append_points({
                HistStat.DO_SWITCH: event.get_event_info(HistStat.DO_SWITCH) * coefficients[HistStat.DO_SWITCH],
                HistStat.DS_SWITCH: event.get_event_info(HistStat.DS_SWITCH) * coefficients[HistStat.DS_SWITCH]
            }, time)

        # Update limits to fit new point
        max_view = max(time, self.pressure_plot.viewRange()[0][1], 10)
        self.limits = (0, max_view)
        self.reset_limits()
        # Update view range iff was already zoomed out
        if currently_zoomed_out:
            # Only need to do pressure plot since plot x ranges are linked
            self.pressure_plot.setXRange(self.limits[0], self.limits[1])

    def reset_limits(self):
        self.pressure_plot.setLimits(xMin=self.limits[0], xMax=self.limits[1], minXRange=10)
        self.slope_plot.setLimits(xMin=self.limits[0], xMax=self.limits[1], minXRange=10)
        self.switch_time_plot.setLimits(xMin=self.limits[0], xMax=self.limits[1], minXRange=10)

    # Assumes list is sorted by time
    def load_event_list(self, event_list):
        for event in event_list:
            self.add_event(event)

        # Only need to do pressure plot since plot x ranges are linked
        self.pressure_plot.setXRange(self.limits[0], self.limits[1])

    def update_settings(self, key):
        if key == "plotting_coefficients":
            plotting_coefficients = self.config_manager.get_settings(key)
            self.coefficients = self.define_coefficients(plotting_coefficients)

    def render_event_time(self, event):
        time = None if event is None else event.event_time - self.initial_time
        self.pressure_plot.add_vertical_line(event.event_type, time)
        self.slope_plot.add_vertical_line(event.event_type, time)
        self.switch_time_plot.add_vertical_line(event.event_type, time)

    def set_sample_sensor(self, connected):
        self.pressure_plot.toggle_line_visibility(HistStat.S_PRESS, connected)
        self.slope_plot.toggle_line_visibility(HistStat.PS_SLOPE, connected)
        self.slope_plot.toggle_line_visibility(HistStat.DS_SLOPE, connected)
        self.switch_time_plot.toggle_line_visibility(HistStat.PS_SWITCH, connected)
        self.switch_time_plot.toggle_line_visibility(HistStat.DS_SWITCH, connected)

    def set_log_coefficients(self, coefficients):
        self.log_coefficients = self.define_coefficients(coefficients)

    # Dictionary of coefficient to apply when plotting each channel
    @staticmethod
    def define_coefficients(plotting_coefficients):
        if plotting_coefficients is None:
            return None
        # Dictionary of coefficient to apply when plotting each channel
        coefficients = {
            HistStat.O_PRESS: plotting_coefficients[Channel.HI_PRE_ORIG],
            HistStat.S_PRESS: plotting_coefficients[Channel.HI_PRE_SAMPLE],
            HistStat.DO_SLOPE: plotting_coefficients[Channel.HI_PRE_ORIG] * 4,
            HistStat.DS_SLOPE: plotting_coefficients[Channel.HI_PRE_SAMPLE] * 4,
            HistStat.PO_SLOPE: plotting_coefficients[Channel.HI_PRE_ORIG] * 4,
            HistStat.PS_SLOPE: plotting_coefficients[Channel.HI_PRE_SAMPLE] * 4,
            HistStat.DO_SWITCH: 0.25,
            HistStat.DS_SWITCH: 0.25,
            HistStat.PO_SWITCH: 0.25,  # ms per index
            HistStat.PS_SWITCH: 0.25,  # ms per index
        }
        return coefficients



from PySide6.QtWidgets import QWidget, QMenu, QScrollArea, QVBoxLayout, QWidgetAction, QLabel
from PySide6.QtCore import Qt, QSize, QPointF, QRectF
from PySide6.QtGui import QFontMetrics, QPainter, QColor


class ScrollableMenu(QMenu):
    def __init__(self, max_messages=100, parent=None):
        super().__init__(parent)

        self.max_messages = max_messages
        self.messages = 0

        # Custom widget to hold the menu actions
        self.container_widget = QWidget()
        self.layout = QVBoxLayout(self.container_widget)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setSpacing(0)
        self.layout.setAlignment(Qt.AlignTop)

        # Scroll area
        self.scroll_area = QScrollArea()
        self.scroll_area.setWidget(self.container_widget)
        self.scroll_area.setWidgetResizable(True)
        self.scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.scroll_area.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)

        # Setting the size of the scroll area
        self.scroll_area.setFixedWidth(900)
        self.scroll_area.setFixedHeight(300)

        # Add the scroll area to the menu
        scroll_area_widget_action = QWidgetAction(self)
        scroll_area_widget_action.setDefaultWidget(self.scroll_area)
        self.addAction(scroll_area_widget_action)

    def add_message(self, message, color):
        label = WrappedLabel(message, color, self)
        label.setStyleSheet("text-align: left")

        # Check if we're scrolled to the top before adding the new message
        scroll_bar = self.scroll_area.verticalScrollBar()
        at_top = scroll_bar.value() <= 5  # Consider "at top" if within 5 pixels

        # Insert the new message at the top
        self.layout.insertWidget(0, label, alignment=Qt.AlignTop)

        self.messages += 1

        # Remove the oldest message if we've exceeded the maximum
        if self.messages > self.max_messages:
            item = self.layout.takeAt(self.layout.count() - 1)
            if item:
                widget = item.widget()
                if widget:
                    widget.deleteLater()
            self.messages -= 1

        # Scroll to the top only if we were already at the top
        if at_top:
            self.scroll_area.verticalScrollBar().setValue(0)
        else:
            # Adjust scroll position to keep the view stable
            new_value = scroll_bar.value() + label.sizeHint().height()
            scroll_bar.setValue(new_value)

    '''
    Returns a list of any widgets in the layout that include the given string
    '''
    def return_matching(self, seq):
        matching = []
        for i in range(self.layout.count()):
            widget=self.layout.itemAt(i).widget()
            if(hasattr(widget,'text')):
                label = widget.text() 
                if(seq in label):
                    matching += [self.layout.itemAt(i)]
        return matching

    '''
    Given a list of widgets will go through and remove each one
    If only one widget is being removed in shoudl be passed in wrapped in []
    '''
    def remove_items(self,items):
        for item in items:
            #This clears them from the scroll menu
            if item:
                widget = item.widget()
                if widget:
                    widget.deleteLater()
            
            #Remove from the layout
            self.layout.removeItem(item)

class WrappedLabel(QLabel):
    def __init__(self, text, color, parent=None):
        super().__init__(text, parent)
        self.setWordWrap(True)
        self.setTextFormat(Qt.PlainText)
        self.color = color

        self.setMinimumWidth(880)  # Adjust as needed
        self.setMaximumWidth(880)  # Adjust as needed
        self.line_indent = 20  # Space for the arrow

    def paintEvent(self, event):
        painter = QPainter(self)
        metrics = QFontMetrics(self.font())
        leading = metrics.leading()
        height = 0
        width = self.width() - self.line_indent  # Adjust width for indent

        for line in self.text().split('\n'):
            if not line:
                height += leading
            is_first_line = True
            while line:
                if is_first_line:
                    indent = 0
                else:
                    indent = self.line_indent
                    # Draw the arrow for wrapped lines
                    painter.setPen(QColor(100, 100, 100))  # Gray color for the arrow
                    painter.drawText(QRectF(2, height, indent - 2, metrics.height()), Qt.AlignLeft | Qt.AlignVCenter,
                                     "↪")

                line = metrics.elidedText(line, Qt.ElideNone, width)
                line_width = metrics.horizontalAdvance(line)
                if line_width <= width:
                    painter.setPen(QColor(self.color))
                    painter.drawText(QPointF(indent, height + metrics.ascent()), line)
                    height += metrics.height()
                    break
                i = len(line)
                while i > 0:
                    i -= 1
                    if metrics.horizontalAdvance(line[:i]) <= width:
                        painter.setPen(QColor(self.color))  # Black color for the text
                        painter.drawText(QPointF(indent, height + metrics.ascent()), line[:i])
                        height += metrics.height()
                        line = line[i:].lstrip()
                        is_first_line = False
                        break
        painter.end()

    def sizeHint(self):
        width = self.width() - self.line_indent
        metrics = QFontMetrics(self.font())
        height = 0
        for line in self.text().split('\n'):
            if not line:
                height += metrics.leading()
            else:
                height += metrics.height() * ((metrics.horizontalAdvance(line) + width - 1) // width)
        return QSize(self.width(), height)



from icarus_v2.backend.event_handler import EventHandler
import numpy as np
from icarus_v2.backend.event import Event, Channel, get_channel


# Detects a period event and transmits data to plot
class PeriodHandler(EventHandler):
    def __init__(self, loader, signal, sample_rate, update_rate, event_report_range) -> None:
        super().__init__(loader, signal, sample_rate, update_rate)
        self.event_report_range = event_report_range # tuple of range of ms around an event to report e.g. (-10,140)
        self.last_depressurize_bit = None # variable to keep track of edges of data chunks in case an event lines up with the start of a chunk
        self.last_depressurize_event = None # variable to keep track of index of last depressurize event
        self.event_type = Event.PERIOD


    # Data: one chunk from the reader
    # Returns whether a depressurize event occurs and the index of the event
    # This logic is the same as for DepressurizeHandler
    def detect_event(self, data):
        depre_valve = get_channel(data, Channel.DEPRE_VALVE)
        # valve array with prior value inserted and last value popped
        valve_offset = np.insert(depre_valve, 0, self.last_depressurize_bit)[:-1]

        # Find indices where there is a low bit preceded by a high bit
        # np.where returns a tuple, so take the first element
        low_indices = np.where(valve_offset & ~depre_valve)[0]

        # All values are high. No event.
        if len(low_indices) == 0:
            event = False
            index = -1

        # Event occurred
        else:
            event = True
            index = low_indices[0]

            # Raise warning if 2 low pulses detected in same chunk. Under default settings this will never happen unless 2 pulses are made within 16ms of each other
            if len(low_indices) > 1:
                raise RuntimeError("Depressurize event dropped on period handler. Two events occured in same data chunk.")

        self.last_depressurize_bit = depre_valve[-1]
        return event, index


    # Returns data to graph. 
    # Always returns from the last event since a period must wait for the second event to get the duration.
    def handle_event(self, event_index):
        # No prior depressurize events
        if self.last_depressurize_event is None:
            data = None
            current_event_idx = None

        else:
            # Update event_report_range based on period length and amount of data displayed before a period
            before, current_after = self.event_report_range
            sample_rate_kHz = float(self.sample_rate) / 1000
            period_duration = float(event_index - self.last_depressurize_event) / sample_rate_kHz # in milliseconds
            after = period_duration - before
            self.event_report_range = (before, after)

            # Get data from last event based on new event_report_range
            data = self.get_event_data(self.last_depressurize_event)
            current_event_idx = int( - before * sample_rate_kHz)

        # Record event index
        self.last_depressurize_event = event_index
        # Return data to plot
        return data, current_event_idx



from __future__ import annotations


def _round_float(number: float, decimal_points: int = 3) -> float:
    decimal = 10**decimal_points
    return round(number * decimal) / decimal


class RGBA:
    """Class handling RGBA color code."""

    def __init__(self, r: int, g: int, b: int, a: float = 1) -> None:
        """Initialize rgba value.

        Args:
            r: Red(0~255).
            g: Green(0~255).
            b: Blue(0~255).
            a: Alpha(0~1). Defaults to 1.
        """
        self._r = min(255, max(0, r)) | 0
        self._g = min(255, max(0, g)) | 0
        self._b = min(255, max(0, b)) | 0
        self._a = _round_float(max(min(1, a), 0))

    def __str__(self) -> str:
        """Format RGBA class.

        e.g. rgba(100, 100, 100, 0.5).
        """
        return f"rgba({self.r}, {self.g}, {self.b}, {self.a:.3f})"

    def __getitem__(self, item: int) -> int | float:
        """Unpack to (r, g, b, a)."""
        return [self.r, self.g, self.b, self.a][item]

    def __eq__(self, other: RGBA) -> bool:
        """Returns true if `r`, `g`, `b` and `a` are all the same."""
        return [self.r, self.g, self.b, self.a] == [other.r, other.g, other.b, other.a]

    @property
    def r(self) -> int:
        return self._r

    @property
    def g(self) -> int:
        return self._g

    @property
    def b(self) -> int:
        return self._b

    @property
    def a(self) -> float:
        return self._a



from threading import Event
import numpy as np

# One producer, N consumer, which each read sequentially from the start without skipping any data.
# Set size buffer composed of an np array.
# Assumes data is only loaded along dimension 0.
# Indexes are stored absolutely, so use head % capacity to get a list index.
class RingBuffer:
    def __init__(self, shape, dtype=int):
        if isinstance(shape, int):
            shape = (shape,)
        self.capacity = shape[0]
        self.buffer = np.empty(shape, dtype=dtype)
        self.write_index = 0
        self.new_data = Event() # Notifies waiting threads when new data has been enqueued


    # Data is copied by reference. Be careful about changing data.
    # Overwrites oldest data
    def enqueue(self, data):
        start = self.write_index % self.capacity
        end = start + len(data)
        if end < self.capacity:
            self.buffer[start:end] = data
        # Case where writing over end of buffer
        else:
            self.buffer[start:] = data[:self.capacity - start]
            self.buffer[:end - self.capacity] = data[self.capacity - start:]
        self.write_index += len(data)

        # Notify readers that new data is available
        self.new_data.set()


# All readers should be terminated before the writer
class SPMCRingBufferReader:
    def __init__(self, buffer):
        self.buffer = buffer
        # Start with nothing to read
        self.read_index = buffer.write_index # Next index of the buffer for this reader to read


    # Returns view of range of data without advancing read_index
    def retrieve_range(self, start, end, timeout=2):
        cap = self.buffer.capacity
        start = max(0, start)

        if end - start > cap:
            raise RuntimeError("Retrieving more data than buffer can store")

        # Raise error if the reader was lapped, meaning some data was overwritten
        if self.buffer.write_index >= end + cap:
            raise RuntimeError(f"Reader was lapped. Reading at {self.read_index} when head is at {self.buffer.write_index}.")

        # Wait for data to be available up till end
        while end > self.buffer.write_index:
            if self.buffer.new_data.wait(timeout):
                self.buffer.new_data.clear()
            else:
                raise TimeoutError

        # Case where range goes through end of buffer
        if end % cap < start % cap:
            a1 = self.buffer.buffer[start % cap:]
            a2 = self.buffer.buffer[:end % cap]
            return np.concatenate((a1,a2))
        else:
            return self.buffer.buffer[start % cap:end % cap]


    # Read block of size size starting at the last index read by this reader
    # Returns a view of that data and the starting index of it
    def read(self, size, timeout=2):
        end = self.read_index + size
        data = self.retrieve_range(self.read_index, end, timeout=timeout)

        self.read_index += size

        return data, self.read_index - size



from icarus_v2.backend.event import Channel, get_channel, Event
from PySide6.QtCore import QObject
from icarus_v2.backend.configuration_manager import ConfigurationManager


# Detects whether the sample sensor is connected to the sample tube by receiving depressurize events
class SampleSensorDetector(QObject):
    low_threshold = 0.998
    high_threshold = 1.05

    def __init__(self, signal=None):
        super().__init__()
        self.config_manager = ConfigurationManager()
        self.config_manager.settings_updated.connect(self.update_settings)
        self.coefficients = self.config_manager.get_settings("plotting_coefficients")
        self.last_result = None

        self.sample_sensor_connected = signal

    def detect(self, event):
        if event.event_type == Event.DEPRESSURIZE:
            if not self.detect_sensor(event):
                # require 2 differences in a row to decrease likelihood of false positives
                if not self.last_result:
                    self.sample_sensor_connected.emit(False)
                self.last_result = False
            else:
                if self.last_result:
                    self.sample_sensor_connected.emit(True)
                self.last_result = True

    def detect_from_list(self, events, coefficients):
        # base decision off last 10 events
        depressurize = []
        for event in events[::-1]:
            if event.event_type == Event.DEPRESSURIZE:
                depressurize.append(event)
            if len(depressurize) == 10:
                break
        num_connected = sum([self.detect_sensor(event, coefficients) for event in depressurize])
        if len(depressurize) > 0:
            sample_sensor_connected = num_connected / len(depressurize) > 0.5
        else:
            sample_sensor_connected = True
        self.sample_sensor_connected.emit(sample_sensor_connected)

    def detect_sensor(self, depressurize_event, coefficients=None):
        if coefficients is None:
            coefficients = self.coefficients

        origin = get_channel(depressurize_event, Channel.HI_PRE_ORIG)[:depressurize_event.event_index].astype('float64')
        sample = get_channel(depressurize_event, Channel.HI_PRE_SAMPLE)[:depressurize_event.event_index].astype('float64')
        origin *= coefficients[Channel.HI_PRE_ORIG]
        sample *= coefficients[Channel.HI_PRE_SAMPLE]

        # AC noise is above 3 kBar, experiments are never run at this pressure
        if sample.max() > 3.0:
            return False

        # prevent division by 0
        origin_mean = origin.mean()
        if origin_mean == 0.0:
            return False

        # percent difference between the two signals
        diff = (origin.mean() - sample.mean()) / origin.mean()

        # If within range, sample sensor is reading a 100% difference and thus is not connected
        if self.low_threshold < diff < self.high_threshold:
            return False
        else:
            return True

    def update_settings(self, key):
        if key == "plotting_coefficients":
            self.coefficients = self.config_manager.get_settings(key)



from PySide6.QtCore import QThread
from time import sleep, time
from icarus_v2.backend.configuration_manager import ConfigurationManager


# DIGITAL
# CH0: high pressure pump
# CH1: depressurize valve
# CH2: pressurize valve
# CH3: spare
# CH4: log
# CH5: spare
# CH6: spare

class PulseGenerator(QThread):
    PUMP = 0
    DEPRESSURIZE = 1
    PRESSURIZE = 2
    LOG = 4

    def __init__(self) -> None:
        super().__init__()

        self.device = None
        self.config_manager = ConfigurationManager()

        # dictionary containing pressurize_width, depressurize_width, period_width, delay_width
        self.settings = self.config_manager.get_settings('timing_settings')
        self.config_manager.settings_updated.connect(self.update_settings)

        # Whether the device should be currently generating pulses
        self.pulsing = False

    def set_device(self, device):
        self.device = device

    # Pressurizes and depressurizes at regular intervals
    def run(self):
        self.pulsing = True
        while self.pulsing:
            # Make sure widths are not edited in the middle of a period.
            # Otherwise, this could cause sleeping for negative times.
            pressurize_width = self.settings["pressurize_width"]
            depressurize_width = self.settings["depressurize_width"]
            period_width = self.settings["period_width"]
            delay_width = self.settings["delay_width"]
            # Get time before setting DIO for more precise timing
            begin_time = time()

            self._pulse_low(self.DEPRESSURIZE, depressurize_width)
            if not self.no_hang_sleep(begin_time + delay_width):
                break

            self._pulse_low(self.PRESSURIZE, pressurize_width)
            if not self.no_hang_sleep(begin_time + period_width):
                break

    # Sleep until end_time, checking for self.pulsing frequently
    # Returns False if self.pulsing becomes false, true otherwise
    def no_hang_sleep(self, end_time, running_check_hz=10):
        remaining_time = end_time - time()
        while remaining_time >= 0:
            sleep_time = min(1 / running_check_hz, remaining_time)
            sleep(sleep_time)
            remaining_time = end_time - time()

            if not self.pulsing:
                return False
        return True

    def set_pressurize_low(self):
        self._set_low(self.PRESSURIZE)

    def set_pressurize_high(self):
        self._set_high(self.PRESSURIZE)

    def set_depressurize_low(self):
        self._set_low(self.DEPRESSURIZE)

    def set_depressurize_high(self):
        self._set_high(self.DEPRESSURIZE)

    def set_pump_low(self):
        self._set_low(self.PUMP)

    def set_pump_high(self):
        self._set_high(self.PUMP)

    def update_settings(self, key = 'timing_settings'):
        if key == 'timing_settings':
            self.settings = self.config_manager.get_settings(key)

    # Sets channel low for duration milliseconds
    # Raises RuntimeError if channel is already low
    def _pulse_low(self, channel, duration):
        if self.device is None:
            raise RuntimeError("Running PulseGenerator when device is not initialized")
        # int representing the current state of dio
        current_dio = self.device.get_current_dio()
        # binary representation of channel to pulse
        channel_bit = 2 ** channel

        # Make sure the channel we are pulsing starts high.
        if not current_dio & channel_bit: # bitwise AND
            raise RuntimeError(f"Error: pulsing low digital channel {channel} which is already low.")

        # Get time before setting DIO for more precise timing
        current_time = time()

        # Set specified channel low
        self.device.set_dio(current_dio ^ channel_bit) # bitwise XOR

        # Sleep for remaining time
        duration_sec = float(duration) / 1000
        time_elapsed = time() - current_time
        remaining_time = max(0, duration_sec - time_elapsed)
        sleep(remaining_time)

        # Reset to original DIO
        self.device.set_dio(current_dio)

    # Sets channel low
    # Raises RuntimeError if channel is already low
    def _set_low(self, channel):
        if self.device is None:
            raise RuntimeError("Running PulseGenerator when device is not initialized")
        # int representing the current state of dio
        current_dio = self.device.get_current_dio()
        # binary representation of channel to pulse
        channel_bit = 2 ** channel

        # Make sure the channel we are setting starts high.
        if not current_dio & channel_bit:  # bitwise AND
            raise RuntimeError(f"Error: setting low digital channel {channel} which is already low.")

        # set low
        self.device.set_dio(current_dio ^ channel_bit)  # bitwise XOR

    # Sets channel high
    # Raises RuntimeError if channel is already high
    def _set_high(self, channel):
        if self.device is None:
            raise RuntimeError("Running PulseGenerator when device is not initialized")
        # int representing the current state of dio
        current_dio = self.device.get_current_dio()
        # binary representation of channel to pulse
        channel_bit = 2 ** channel

        # Make sure the channel we are setting starts low.
        if current_dio & channel_bit:  # bitwise AND
            raise RuntimeError(f"Error: setting high digital channel {channel} which is already high.")

        # set high
        self.device.set_dio(current_dio | channel_bit)  # bitwise OR

    def quit(self):
        self.pulsing = False



import importlib
from icarus_v2.backend.configuration_manager import ConfigurationManager
from icarus_v2.gui.main_window import MainWindow
from icarus_v2.backend.data_handler import DataHandler
from PySide6.QtWidgets import QApplication
from PySide6.QtGui import QIcon


# Application entry point
def main():
    app = QApplication([])
    app.setApplicationName('Icarus')

    with importlib.resources.path('icarus_v2.resources', 'wing.png') as icon_path:
        app.setWindowIcon(QIcon(str(icon_path)))

    # initialize singleton ConfigurationManager
    ConfigurationManager()

    # initialize gui
    window = MainWindow()
    window.showMaximized()

    # start data collection
    data_handler = DataHandler()
    window.set_device(data_handler)
    data_handler.start()

    app.exec()


if __name__ == "__main__":
    main()



import numpy as np
from icarus_v2.backend.event import Event, Channel, get_channel
from PySide6.QtWidgets import QGridLayout, QWidget
from icarus_v2.gui.styled_plot_widget import StyledPlotWidget
from icarus_v2.backend.configuration_manager import ConfigurationManager


class EventPlot(QWidget):
    DISPLAY_CHANNELS = {
        Event.PRESSURIZE: [
            Channel.TARGET,
            Channel.HI_PRE_SAMPLE,
            Channel.HI_PRE_ORIG,
            Channel.DEPRE_VALVE,
            Channel.PRE_VALVE,
            Channel.PRE_LOW,
            Channel.PRE_UP
        ],
        Event.DEPRESSURIZE: [
            Channel.TARGET,
            Channel.HI_PRE_SAMPLE,
            Channel.HI_PRE_ORIG,
            Channel.DEPRE_VALVE,
            Channel.PRE_VALVE,
            Channel.DEPRE_LOW,
            Channel.DEPRE_UP
        ],
        Event.PERIOD: [
            Channel.TARGET,
            Channel.HI_PRE_SAMPLE,
            Channel.HI_PRE_ORIG,
            Channel.DEPRE_VALVE,
            Channel.PRE_VALVE
        ]
    }

    def __init__(self, event_type, parent=None):
        super().__init__(parent=parent)

        self.event_type = event_type
        self.config_manager = ConfigurationManager()
        self.config_manager.settings_updated.connect(self.update_settings)
        self.coefficients = self.config_manager.get_settings("plotting_coefficients")
        self.log_coefficients = None
        self.hide_valve_setting = self.config_manager.get_settings("hide_valve_sensors")

        self.plot = StyledPlotWidget()
        self.plot.setYRange(0, 3)
        self.plot.set_y_label('Pressure (kbar)')

        for channel in self.DISPLAY_CHANNELS[event_type]:
            self.plot.add_line(channel)

        # local statistics functions
        def get_valve_open_time(y_data, srate):
            low_idx = y_data.argmin()
            duration = y_data[low_idx:].argmax()

            # Case where the end of the event is not in the reported data
            if y_data[low_idx + duration] == 0:
                duration = len(y_data) - low_idx

            return duration / srate

        def get_period_width(y_data, srate):
            """
            Periods are defined as the distance between depressurize signals.
            Period events have buffers of 10ms on either side.
            This is a magic number, but it's way easier since period events are down-sampled.
            """
            duration = len(y_data) / srate
            duration -= 10 * 2 / 1000

            return duration

        def get_delay_width(y_data, srate):
            """
            Delay is the time from the depressurize event to pressurize.
            There is a 10ms buffer on the left side of the depressurize event.
            """
            duration = np.argmin(y_data) / srate
            duration -= 10 / 1000

            return  duration

        if event_type == Event.PRESSURIZE:
            self.x_unit = 'ms'
            self.plot.set_title("Pressurize")
            self.plot.add_statistic(Channel.PRE_VALVE, get_valve_open_time, "Valve Open (ms): {:.2f}")
        elif event_type == Event.DEPRESSURIZE:
            self.x_unit = 'ms'
            self.plot.set_title("Depressurize")
            self.plot.add_statistic(Channel.DEPRE_VALVE, get_valve_open_time, "Valve Open (ms): {:.2f}")
        else:
            self.x_unit = 's'
            self.plot.set_title("Period")
            self.plot.add_statistic(Channel.DEPRE_VALVE, get_period_width, "Period (s): {:.2f}")
            self.plot.add_statistic(Channel.PRE_VALVE, get_delay_width, "Delay (s): {:.2f}:")

        self.plot.set_x_label(f'Time ({self.x_unit})')
        self.hide_valve_sensors(self.hide_valve_setting)

        # Set layout
        layout = QGridLayout()
        layout.addWidget(self.plot)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

    def update_data(self, event):
        if event is None:
            self.plot.reset()
            return

        data = event.data
        # Calculate times based on event.step_time and event_index
        frequency = event.step_time
        if self.x_unit == "s":  # frequencies are in ms by default
            frequency /= 1000
        srate = 1.0 / frequency

        time_before_event = - event.event_index * frequency
        time_after_event = (len(event.data) - event.event_index - 1) * frequency
        times = np.linspace(time_before_event, time_after_event, len(data))

        # update data for each line
        coefficients = self.coefficients if self.log_coefficients is None else self.log_coefficients
        for channel in self.DISPLAY_CHANNELS[self.event_type]:
            y_data = get_channel(data, channel) * coefficients[channel]
            self.plot.update_line_data(channel, times, y_data, srate)

    def update_settings(self, key):
        if key == "plotting_coefficients":
            self.coefficients = self.config_manager.get_settings(key)
        elif key == "hide_valve_sensors":
            self.hide_valve_setting = self.config_manager.get_settings(key)
            self.hide_valve_sensors(self.hide_valve_setting)

    def hide_valve_sensors(self, hide_valve_sensors):
        valve_sensors = [Channel.PRE_LOW, Channel.PRE_UP, Channel.DEPRE_LOW, Channel.DEPRE_UP]
        visible = not hide_valve_sensors

        for channel in valve_sensors:
            self.plot.toggle_line_visibility(channel, visible)

    # Hide the sample sensor if it is disconnected.
    def set_sample_sensor(self, connected):
        self.plot.toggle_line_visibility(Channel.HI_PRE_SAMPLE, connected)

    def set_log_coefficients(self, coefficients):
        self.log_coefficients = coefficients



from PySide6.QtWidgets import (
    QGridLayout,
    QGroupBox,
    QLabel,
    QSizePolicy
)
from icarus_v2.backend.event import Channel
from icarus_v2.backend.configuration_manager import ConfigurationManager


# Panel for pressure
class PressureDisplay(QGroupBox):
    def __init__(self):
        super().__init__()

        self.config_manager = ConfigurationManager()
        self.coefficients = self.config_manager.get_settings("plotting_coefficients")
        self.config_manager.settings_updated.connect(self.update_settings)

        # Displays
        self.target_display = QLabel("0.000")
        self.origin_display = QLabel("0.000")

        # Labels
        main_label = QLabel("Pressure(kbar):")
        target_label = QLabel("Target:")
        origin_label = QLabel("Origin:")

        # Set main layout
        layout = QGridLayout()
        layout.setSpacing(0)
        layout.addWidget(main_label, 0, 0, 1, 0)
        layout.addWidget(target_label, 1, 0)
        layout.addWidget(origin_label, 2, 0)
        layout.addWidget(self.target_display, 1, 1)
        layout.addWidget(self.origin_display, 2, 1)

        self.setStyleSheet("font-size: 28pt;")
        self.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Minimum)
        self.setLayout(layout)


    def update_pressure(self, event):
        target_pressure = event.get_target_pressure() * self.coefficients[Channel.TARGET]
        origin_pressure = event.get_origin_pressure() * self.coefficients[Channel.HI_PRE_ORIG]
        self.target_display.setText(f"{target_pressure:.3f}")
        self.origin_display.setText(f"{origin_pressure:.3f}")


    def reset(self):
        self.target_display.setText("0.000")
        self.origin_display.setText("0.000")

    def update_settings(self, key):
        if key == "plotting_coefficients":
            self.coefficients = self.config_manager.get_settings(key)



from PySide6.QtWidgets import (
    QDialog,
    QDialogButtonBox,
    QGridLayout,
    QLabel,
    QLineEdit,
    QVBoxLayout,
    QGroupBox,
    QPushButton,
    QCheckBox,
    QSpacerItem,
    QSizePolicy,
    QWidget,
    QStackedWidget,
    QToolBar,
    QComboBox
)
from PySide6.QtGui import QDoubleValidator, QIntValidator, QAction
from icarus_v2.gui.error_dialog import open_error_dialog
from icarus_v2.backend.event import Channel, get_channel
import numpy as np
from PySide6.QtCore import Qt
from icarus_v2.backend.configuration_manager import ConfigurationManager


class SettingsDialog(QDialog):
    def __init__(self, connected, pressure_signal, sentry, parent=None):
        super().__init__(parent=parent)

        self.config_manager = ConfigurationManager()
        self.pressure_signal = pressure_signal
        self.sentry = sentry

        # Create the toolbar
        toolbar = QToolBar(parent)
        action1 = QAction("General", self)
        action2 = QAction("Advanced", self)
        action1.triggered.connect(self.show_general)
        action2.triggered.connect(self.show_advanced)
        toolbar.addAction(action1)
        toolbar.addAction(action2)

        self.general_widget = self.get_default_widget(connected)
        self.advanced_widget = self.get_advanced_widget(connected)
        self.stack = QStackedWidget()
        self.stack.addWidget(self.general_widget)
        self.stack.addWidget(self.advanced_widget)

        # Button to apply changes
        buttons = QDialogButtonBox.Apply
        button_box = QDialogButtonBox(buttons)
        button_box.button(QDialogButtonBox.Apply).clicked.connect(self.apply)

        # Dialog configuration
        self.setWindowTitle("Settings")
        self.setStyleSheet("font-size: 17pt;")
        self.setMinimumWidth(610)

        # Main layout for the dialog
        main_layout = QVBoxLayout(self)
        main_layout.addWidget(toolbar)
        main_layout.addWidget(self.stack)
        main_layout.addWidget(button_box)

    def get_default_widget(self, connected):
        # Allow only positive floating-point numbers
        positive_float = QDoubleValidator()
        positive_int = QIntValidator()
        positive_float.setBottom(0)
        positive_int.setBottom(0)
        # Width of edit boxes
        edit_width = 190

        # View section
        self.hide_valve_checkbox = QCheckBox()
        hide_valve = self.config_manager.get_settings('hide_valve_sensors')
        self.hide_valve_checkbox.setCheckState(Qt.Checked if hide_valve else Qt.Unchecked)
        self.hide_valve_checkbox.stateChanged.connect(self.set_hide_valve)
        self.theme_combo_box = QComboBox()
        self.theme_combo_box.addItems(["Dark", "Light"])
        self.theme_combo_box.setFixedWidth(edit_width)
        theme = self.config_manager.get_settings('theme')
        self.theme_combo_box.setCurrentText(theme.capitalize())
        self.theme_combo_box.currentTextChanged.connect(self.set_theme)

        view_layout = QGridLayout()
        view_layout.addWidget(QLabel("Theme:"), 0, 0)
        view_layout.addItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Preferred), 0, 1)
        view_layout.addWidget(self.theme_combo_box, 0, 2)
        view_layout.addWidget(QLabel("Hide Valve Sensors:"), 1, 0)
        view_layout.addItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Preferred), 1, 1)
        view_layout.addWidget(self.hide_valve_checkbox, 1, 2, alignment=Qt.AlignRight)
        view_group = QGroupBox("View")
        view_group.setLayout(view_layout)

        # Pulse timings section
        self.pressurize_width_edit = QLineEdit()
        self.depressurize_width_edit = QLineEdit()
        self.period_width_edit = QLineEdit()
        self.delay_width_edit = QLineEdit()
        self.pressurize_width_edit.setFixedWidth(edit_width)
        self.depressurize_width_edit.setFixedWidth(edit_width)
        self.period_width_edit.setFixedWidth(edit_width)
        self.delay_width_edit.setFixedWidth(edit_width)
        self.pressurize_width_edit.setValidator(positive_float)
        self.depressurize_width_edit.setValidator(positive_float)
        self.period_width_edit.setValidator(positive_float)
        self.delay_width_edit.setValidator(positive_float)
        self.coefficients = self.config_manager.get_settings('plotting_coefficients')
        self.timing_settings = self.config_manager.get_settings('timing_settings')
        self.pressurize_width_edit.setText(str(self.timing_settings['pressurize_width']))
        self.depressurize_width_edit.setText(str(self.timing_settings['depressurize_width']))
        self.period_width_edit.setText(str(self.timing_settings['period_width']))
        self.delay_width_edit.setText(str(self.timing_settings['delay_width']))
        self.pressurize_width_edit.textChanged.connect(self.set_pressurize_width)
        self.depressurize_width_edit.textChanged.connect(self.set_depressurize_width)
        self.period_width_edit.textChanged.connect(self.set_period_width)
        self.delay_width_edit.textChanged.connect(self.set_delay_width)

        timings_layout = QGridLayout()
        timings_layout.addWidget(QLabel("Pressurize Width (ms):"), 0, 0)
        timings_layout.addWidget(QLabel("Depressurize Width (ms):"), 1, 0)
        timings_layout.addWidget(QLabel("Period (s):"), 2, 0)
        timings_layout.addWidget(QLabel("Delay (s):"), 3, 0)
        timings_layout.addWidget(self.pressurize_width_edit, 0, 1)
        timings_layout.addWidget(self.depressurize_width_edit, 1, 1)
        timings_layout.addWidget(self.period_width_edit, 2, 1)
        timings_layout.addWidget(self.delay_width_edit, 3, 1)
        timings_group = QGroupBox("Pulse Timings")
        timings_group.setLayout(timings_layout)

        # Hardware section
        self.pump_count_edit = QLineEdit()
        self.pressurize_count_edit = QLineEdit()
        self.depressurize_count_edit = QLineEdit()
        self.recalibrate_button = QPushButton("Recalibrate")
        self.recalibrate_button.setEnabled(connected)
        self.pump_count_edit.setFixedWidth(edit_width)
        self.pressurize_count_edit.setFixedWidth(edit_width)
        self.depressurize_count_edit.setFixedWidth(edit_width)
        self.pump_count_edit.setValidator(positive_int)
        self.pressurize_count_edit.setValidator(positive_int)
        self.depressurize_count_edit.setValidator(positive_int)
        self.counter_settings = self.config_manager.get_settings('counter_settings')
        self.pump_count_edit.setText(str(self.counter_settings['pump_count']))
        self.pressurize_count_edit.setText(str(self.counter_settings['pressurize_count']))
        self.depressurize_count_edit.setText(str(self.counter_settings['depressurize_count']))
        self.pump_count_edit.textChanged.connect(self.set_pump_count)
        self.pressurize_count_edit.textChanged.connect(self.set_pressurize_count)
        self.depressurize_count_edit.textChanged.connect(self.set_depressurize_count)
        self.recalibrate_button.clicked.connect(self.get_recalibration)

        hardware_layout = QGridLayout()
        hardware_layout.addWidget(QLabel("Pump Count:"), 0, 0)
        hardware_layout.addWidget(QLabel("Pressurize Count:"), 1, 0)
        hardware_layout.addWidget(QLabel("Depressurize Count"), 2, 0)
        hardware_layout.addWidget(QLabel("Recalibrate Sensors:"), 3, 0)
        hardware_layout.addWidget(self.pump_count_edit, 0, 1)
        hardware_layout.addWidget(self.pressurize_count_edit, 1, 1)
        hardware_layout.addWidget(self.depressurize_count_edit, 2, 1)
        hardware_layout.addWidget(self.recalibrate_button, 3, 1)
        hardware_group = QGroupBox("Hardware")
        hardware_group.setLayout(hardware_layout)

        widget = QWidget()
        layout = QVBoxLayout(widget)
        layout.addWidget(view_group)
        layout.addWidget(timings_group)
        layout.addWidget(hardware_group)
        layout.addItem(QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding))
        return widget

    def get_advanced_widget(self, connected):
        # Allow only positive floating-point numbers
        positive_float = QDoubleValidator()
        positive_int = QIntValidator()
        positive_float.setBottom(0)
        positive_int.setBottom(1)
        # Width of edit boxes
        edit_width = 190

        self.depress_edit = QLineEdit()
        self.press_edit = QLineEdit()
        self.pump_rate_edit = QLineEdit()
        self.example_edit = QLineEdit()
        self.diff_count_error = QLineEdit()
        self.max_pump_edit = QLineEdit()
        self.pump_window_edit = QLineEdit()
        self.enable_sentry_button = QPushButton("Enable")
        self.enable_sentry_button.setEnabled(connected)

        self.depress_edit.setFixedWidth(edit_width)
        self.press_edit.setFixedWidth(edit_width)
        self.pump_rate_edit.setFixedWidth(edit_width)
        self.example_edit.setFixedWidth(edit_width)
        self.diff_count_error.setFixedWidth(edit_width)
        self.pump_window_edit.setFixedWidth(edit_width)
        self.max_pump_edit.setFixedWidth(edit_width)

        self.depress_edit.setValidator(positive_float)
        self.press_edit.setValidator(positive_float)
        self.pump_rate_edit.setValidator(positive_float)
        self.example_edit.setValidator(positive_int)
        self.diff_count_error.setValidator(positive_int)
        self.pump_window_edit.setValidator(positive_float)
        self.max_pump_edit.setValidator(positive_int)

        self.sentry_settings = self.config_manager.get_settings('sentry_settings')

        self.depress_edit.setText(str(self.sentry_settings['max_pressure_before_depress_decrease'] * 100))
        self.press_edit.setText(str(self.sentry_settings['max_pressure_before_depress_increase'] * 100))
        self.pump_rate_edit.setText(str(self.sentry_settings['max_pump_rate_increase'] * 100))
        self.example_edit.setText(str(self.sentry_settings['example_events']))
        self.diff_count_error.setText(str(self.sentry_settings['decrease_count_to_error']))
        self.max_pump_edit.setText(str(self.sentry_settings['max_pumps_in_window']))
        self.pump_window_edit.setText(str(self.sentry_settings['pump_window']))

        self.depress_edit.textChanged.connect(self.set_sentry_depress)
        self.press_edit.textChanged.connect(self.set_sentry_press)
        self.pump_rate_edit.textChanged.connect(self.set_sentry_pump_rate)
        self.example_edit.textChanged.connect(self.set_sentry_example)
        self.diff_count_error.textChanged.connect(self.set_sentry_diff_count)
        self.max_pump_edit.textChanged.connect(self.set_sentry_max_pumps)
        self.pump_window_edit.textChanged.connect(self.set_sentry_pump_window)
        self.enable_sentry_button.clicked.connect(self.enable_sentry)

        error_layout = QGridLayout()
        error_layout.addWidget(QLabel("Pressure Decrease to Error Count:"), 0, 0)
        error_layout.addWidget(QLabel("Max Pump Count in Window:"), 1, 0)
        error_layout.addWidget(QLabel("Pump Window (s):"), 2, 0)
        error_layout.addWidget(self.diff_count_error, 0, 1)
        error_layout.addWidget(self.max_pump_edit, 1, 1)
        error_layout.addWidget(self.pump_window_edit, 2, 1)
        error_group = QGroupBox("Errors")
        error_group.setLayout(error_layout)

        warning_layout = QGridLayout()
        warning_layout.addWidget(QLabel("Max Pressure Decrease (%):"), 0, 0)
        warning_layout.addWidget(QLabel("Max Pressure Increase (%):"), 1, 0)
        warning_layout.addWidget(QLabel("Max Pump Rate Increase (%):"), 2, 0)
        warning_layout.addWidget(self.depress_edit, 0, 1)
        warning_layout.addWidget(self.press_edit, 1, 1)
        warning_layout.addWidget(self.pump_rate_edit, 2, 1)
        warning_group = QGroupBox("Warnings")
        warning_group.setLayout(warning_layout)

        misc_layout = QGridLayout()
        misc_layout.addWidget(QLabel("Example Event Count:"), 0, 0)
        misc_layout.addWidget(QLabel("Enable Sentry (until device disconnected):"), 1, 0)
        misc_layout.addWidget(self.example_edit, 0, 1)
        misc_layout.addWidget(self.enable_sentry_button, 1, 1)
        misc_group = QGroupBox("Sentry")
        misc_group.setLayout(misc_layout)

        widget = QWidget()
        layout = QVBoxLayout(widget)
        layout.addWidget(error_group)
        layout.addWidget(warning_group)
        layout.addWidget(misc_group)
        layout.addItem(QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding))

        return widget

    def show_general(self):
        self.stack.setCurrentIndex(0)

    def show_advanced(self):
        self.stack.setCurrentIndex(1)

    # Apply changes to config_manager and close the dialog
    # Also does some minor error checking for timing settings
    def apply(self) -> None:
        error = None
        if self.timing_settings['pressurize_width'] < 8:
            error = "Pressurize width should be greater than 8ms."
        elif self.timing_settings['depressurize_width'] < 8:
            error = "Depressurize width should be greater than 8ms."
        elif self.timing_settings['period_width'] < 3:
            error = "Period width should be greater than 3s."
        elif self.timing_settings['delay_width'] < 0.05:
            error = "Delay width should be greater than 0.05s."
        elif self.timing_settings['pressurize_width'] / 1000 >= self.timing_settings['period_width']:
            error = "Pressurize width should be less than Period width."
        elif self.timing_settings['depressurize_width'] / 1000 >= self.timing_settings['period_width']:
            error = "Depressurize width should be less than Period width."
        elif self.timing_settings['delay_width'] >= self.timing_settings['period_width']:
            error = "Delay width should be less than Period width."
        elif self.sentry_settings['example_events'] <= 1:
            error = "Example events should be greater than 1."
        elif self.sentry_settings['decrease_count_to_error'] <= 0:
            error = "Difference to error count should be greater than 0."

        if error is not None:
            self.dialog = open_error_dialog(error, QDialogButtonBox.Ok, self)
            # Do not apply changes or close window
            return

        self.config_manager.save_settings("timing_settings", self.timing_settings)
        self.config_manager.save_settings("counter_settings", self.counter_settings)
        self.config_manager.save_settings("sentry_settings", self.sentry_settings)
        self.close()

    def set_pressurize_width(self, pressurize_width):
        try:
            pressurize_width = float(pressurize_width)
        except:
            return
        self.timing_settings['pressurize_width'] = pressurize_width

    def set_depressurize_width(self, depressurize_width):
        try:
            depressurize_width = float(depressurize_width)
        except:
            return
        self.timing_settings['depressurize_width'] = depressurize_width

    def set_period_width(self, period_width):
        try:
            period_width = float(period_width)
        except:
            return
        self.timing_settings['period_width'] = period_width

    def set_delay_width(self, delay_width):
        try:
            delay_width = float(delay_width)
        except:
            return
        self.timing_settings['delay_width'] = delay_width

    def set_pressurize_count(self, pressurize_count):
        try:
            pressurize_count = int(pressurize_count)
        except:
            return
        self.counter_settings['pressurize_count'] = pressurize_count

    def set_depressurize_count(self, depressurize_count):
        try:
            depressurize_count = int(depressurize_count)
        except:
            return
        self.counter_settings['depressurize_count'] = depressurize_count

    def set_pump_count(self, pump_count):
        try:
            pump_count = int(pump_count)
        except:
            return
        self.counter_settings['pump_count'] = pump_count

    # Make target and sample match origin
    def get_recalibration(self):
        def recalibrate(event):
            origin_avg = np.mean(get_channel(event, Channel.HI_PRE_ORIG))
            sample_avg = np.mean(get_channel(event, Channel.HI_PRE_SAMPLE))
            target_avg = np.mean(get_channel(event, Channel.TARGET))

            origin_pressure = origin_avg * self.coefficients[Channel.HI_PRE_ORIG]
            sample_coefficient = origin_pressure / sample_avg
            target_coefficient = origin_pressure / target_avg

            self.coefficients[Channel.HI_PRE_SAMPLE] = sample_coefficient
            self.coefficients[Channel.TARGET] = target_coefficient

            self.config_manager.save_settings('plotting_coefficients', self.coefficients)

        # Send pressure readings exactly once
        self.pressure_signal.connect(recalibrate, type=Qt.SingleShotConnection)

    def set_theme(self, theme):
        self.config_manager.save_settings('theme', theme.lower())

    def set_hide_valve(self, state):
        self.config_manager.save_settings('hide_valve_sensors', bool(state))

    def set_connected(self, connected):
        self.recalibrate_button.setEnabled(connected)
        self.enable_sentry_button.setEnabled(connected)

    def set_pressure_signal(self, pressure_signal):
        self.pressure_signal = pressure_signal

    def set_sentry(self, sentry):
        self.sentry = sentry

    def set_sentry_depress(self, depress):
        try:
            depress = float(depress) / 100
        except:
            return
        self.sentry_settings['max_pressure_before_depress_decrease'] = depress

    def set_sentry_press(self, press):
        try:
            press = float(press) / 100
        except:
            return
        self.sentry_settings['max_pressure_before_depress_increase'] = press

    def set_sentry_pump_rate(self, pump_rate):
        try:
            pump_rate = float(pump_rate) / 100
        except:
            return
        self.sentry_settings['max_pump_rate_increase'] = pump_rate

    def set_sentry_example(self, example):
        try:
            example = int(example)
        except:
            return
        self.sentry_settings['example_events'] = example

    def set_sentry_diff_count(self, diff_count):
        try:
            diff_count = int(diff_count)
        except:
            return
        self.sentry_settings['decrease_count_to_error'] = diff_count

    def set_sentry_pump_window(self, pump_window):
        try:
            pump_window = float(pump_window)
        except:
            return
        self.sentry_settings['pump_window'] = pump_window

    def set_sentry_max_pumps(self, max_pumps):
        try:
            max_pumps = int(max_pumps)
        except:
            return
        self.sentry_settings['max_pumps_in_window'] = max_pumps

    def enable_sentry(self):
        self.sentry.handle_experiment(False)



from PySide6.QtWidgets import (
    QGroupBox,
    QHBoxLayout,
    QVBoxLayout,
    QLabel,
    QPushButton,
    QLineEdit,
    QDialog,
    QFileDialog,
    QSizePolicy,
    QSpacerItem,
    QMessageBox
)
import os
from PySide6.QtGui import QDoubleValidator, QFontMetrics
from PySide6.QtCore import Qt, Signal, QStandardPaths
from icarus_v2.gui.error_dialog import open_error_dialog
from bisect import bisect_right
from icarus_v2.backend.event import Event
from icarus_v2.backend.log_reader import LogReader
from math import ceil
from icarus_v2.backend.sample_sensor_detector import SampleSensorDetector


# Control panel for logs
class LogControlPanel(QGroupBox):
    pressurize_event_signal = Signal(Event)
    depressurize_event_signal = Signal(Event)
    period_event_signal = Signal(Event)
    event_list_signal = Signal(list)
    reset_history_signal = Signal()
    sample_sensor_connected = Signal(bool)
    log_coefficients_signal = Signal(object)

    def __init__(self, parent=None, tool_bar=None):
        super().__init__(parent=parent)

        self.log_reader = LogReader(tool_bar=tool_bar)
        self.press_index = None
        self.depress_index = None
        self.period_index = None
        self.filename = None
        self.edit_dialog = None
        self.name_edit = None
        self.currently_logging = None
        self.sample_sensor_detector = SampleSensorDetector(self.sample_sensor_connected)

        # Title
        title = QLabel("Viewing Log")
        title.setStyleSheet("font-size: 28pt;")
        self.filename_label = QLabel(" ")
        self.filename_label.setFixedHeight(23)

        # Time controls
        self.time_edit = QLineEdit(self)
        self.last_button = QPushButton("Previous")
        self.next_button = QPushButton("Next")
        self.time_edit.editingFinished.connect(self.select_time)
        self.last_button.pressed.connect(self.emit_last_event)
        self.next_button.pressed.connect(self.emit_next_event)

        # File control
        open_button = QPushButton("Open File")
        self.current_button = QPushButton("Open Current")
        self.new_log_button = QPushButton("Start New Log")
        self.edit_button = QPushButton("Edit File")
        open_button.clicked.connect(self.choose_log)
        self.current_button.clicked.connect(self.open_current)
        self.edit_button.clicked.connect(self.edit_file)
        self.new_log_button.clicked.connect(self.new_log)

        # Prefer height of buttons on device control panel
        self.last_button.setMaximumHeight(74)
        self.next_button.setMaximumHeight(74)
        self.current_button.setMaximumHeight(74)
        open_button.setMaximumHeight(74)
        self.edit_button.setMaximumHeight(74)
        self.new_log_button.setMaximumHeight(74)
        self.last_button.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.next_button.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.current_button.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        open_button.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.edit_button.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.new_log_button.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        # Set layout
        time_layout = QHBoxLayout()
        time_layout.addWidget(QLabel("Time (s):"))
        time_layout.addWidget(self.time_edit)

        step_layout = QHBoxLayout()
        step_layout.addWidget(self.last_button)
        step_layout.addWidget(self.next_button)

        layout = QVBoxLayout(self)
        layout.addWidget(title, alignment=Qt.AlignHCenter)
        layout.addWidget(self.filename_label)
        # Spacer to separate sections
        layout.addItem(QSpacerItem(0, 0, QSizePolicy.Preferred, QSizePolicy.Expanding))
        layout.addLayout(time_layout)
        layout.addLayout(step_layout)
        # Spacer to separate sections
        layout.addItem(QSpacerItem(0, 0, QSizePolicy.Preferred, QSizePolicy.Expanding))
        layout.addWidget(self.current_button)
        layout.addWidget(open_button)
        layout.addWidget(self.new_log_button)
        layout.addWidget(self.edit_button)

        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)  # Width of device control panel

        self.set_logging(False)
        self.reset()

    def open_current(self):
        if not self.currently_logging:
            return
        filename = self.log_reader.logger.filename
        self.open_log(filename)

    def choose_log(self):
        log_path = os.path.join(QStandardPaths.writableLocation(QStandardPaths.AppDataLocation), 'logs')

        if not os.path.exists(log_path):
            os.makedirs(log_path)

        file = QFileDialog.getOpenFileName(self, "Open File", log_path, "Log Files (*.xz)")[0]
        # No file selected
        if file == "":
            return
        self.open_log(file)

    def open_log(self, file):
        self.reset_history_signal.emit()
        try:
            self.log_reader.read_events(file)
        except Exception as e:
            self.dialog = open_error_dialog("Incorrect format for log file.")
            return

        self.time_edit.setText("")
        self.next_button.setEnabled(True)
        self.last_button.setEnabled(True)
        self.edit_button.setEnabled(True)
        self.time_edit.setEnabled(True)

        self.press_index = -1
        self.depress_index = -1
        self.period_index = -1

        self.filename = self.log_reader.filename
        self.filename_label.setText(os.path.basename(self.filename))
        # Adjust font size to fit in the view
        for i in range(17, 8, -1):
            self.filename_label.setStyleSheet(f"font-size: {i}px")
            if (QFontMetrics(self.filename_label.font()).boundingRect(os.path.basename(self.filename)).width()
                    < self.width() - 25):
                break

        # Set limit on time line_edit
        if len(self.log_reader.events) > 0:
            upper_bound = self.log_reader.events[-1].event_time - self.log_reader.events[0].event_time
        else:
            upper_bound = 0
        self.time_edit.setValidator(QDoubleValidator(0, upper_bound, 2))

        self.sample_sensor_detector.detect_from_list(
            self.log_reader.events,
            self.log_reader.log_coefficients
        )
        self.log_coefficients_signal.emit(self.log_reader.log_coefficients)
        self.event_list_signal.emit(self.log_reader.events)

    def edit_file(self):
        self.edit_dialog = QDialog(self)

        self.name_edit = QLineEdit()
        basename = os.path.basename(self.filename)
        self.name_edit.setText(basename)
        extension_length = len(basename.split('.')[-1]) + 1
        self.name_edit.setSelection(0, len(basename) - extension_length)

        button_rename = QPushButton("Rename File")
        button_delete = QPushButton("Delete File")
        button_rename.clicked.connect(self.rename_file)
        button_delete.clicked.connect(self.delete_file)

        layout = QVBoxLayout()
        layout.addWidget(self.name_edit)
        layout.addWidget(button_rename)
        layout.addWidget(button_delete)

        self.edit_dialog.setWindowTitle("Edit File")
        self.edit_dialog.setFixedSize(500, 200)
        self.edit_dialog.setLayout(layout)
        self.edit_dialog.show()

    def rename_file(self):
        new_name = self.name_edit.text()
        if new_name:
            new_filename = os.path.join(os.path.dirname(self.filename), new_name)
            try:
                os.rename(self.filename, new_filename)  

                # Update logger filename if this is the currently written log
                if self.currently_logging:
                    if self.filename == self.log_reader.logger.filename:
                        self.log_reader.logger.filename = new_filename 
                self.filename_label.setText(os.path.basename(new_filename))  # Update the label
                self.filename = new_filename
            except OSError as e:
                self.dialog = open_error_dialog(e)
        self.edit_dialog.close()

    def delete_file(self):
        confirm_dialog = QMessageBox.question(self, "Are you sure?", "This action cannot be undone.", QMessageBox.Yes | QMessageBox.Cancel)
        if confirm_dialog == QMessageBox.Yes:
            # Start new log if this is the currently written log
            if self.currently_logging:
                if self.filename == self.log_reader.logger.filename:
                    self.log_reader.logger.new_log_file(self.log_reader.logger.current_path)

            if os.path.exists(self.filename):
                try:
                    os.remove(self.filename)
                except OSError as e:
                    self.dialog = open_error_dialog(e)
                    return

            self.reset()
            self.reset_history_signal.emit()
        self.edit_dialog.close()

    def select_time(self):
        try:
            time = float(self.time_edit.text())
        except Exception as e:
            return
        if len(self.log_reader.events) == 0:
            return

        initial_time = self.log_reader.events[0].event_time
        index = max(0, bisect_right(self.log_reader.events, time, key=lambda x: x.event_time - initial_time) - 1)

        press = None
        depress = None
        period = None
        for i in range(index, -1, -1):
            event = self.log_reader.events[i]
            if press is None and event.event_type == Event.PRESSURIZE:
                self.press_index = i
                press = event
            elif depress is None and event.event_type == Event.DEPRESSURIZE:
                self.depress_index = i
                depress = event
            elif period is None and event.event_type == Event.PERIOD:
                self.period_index = index
                period = event

            if press is not None and depress is not None and period is not None:
                break

        self.pressurize_event_signal.emit(press)
        self.depressurize_event_signal.emit(depress)
        self.period_event_signal.emit(period)

    def emit_next_event(self):
        index = max(self.press_index, self.depress_index, self.period_index) + 1

        while index + 1 < len(self.log_reader.events) and self.log_reader.events[index].event_type == Event.PRESSURE:
            index += 1

        event = self.log_reader.events[index]

        if index >= len(self.log_reader.events):
            return

        if event.event_type == Event.PRESSURIZE:
            self.press_index = index
            self.pressurize_event_signal.emit(event)

        elif event.event_type == Event.DEPRESSURIZE:
            self.depress_index = index
            self.depressurize_event_signal.emit(event)
            # Also emit matching PERIOD
            if index + 1 < len(self.log_reader.events):
                n_event = self.log_reader.events[index + 1]
                if n_event.event_type == Event.PERIOD and abs(n_event.event_time - event.event_time) < 2:
                    self.period_index = index + 1
                    self.period_event_signal.emit(n_event)

        elif event.event_type == Event.PERIOD:
            self.period_index = index
            self.period_event_signal.emit(event)
            # Also emit matching DEPRESSURIZE
            if index + 1 < len(self.log_reader.events):
                n_event = self.log_reader.events[index + 1]
                if n_event.event_type == Event.DEPRESSURIZE and abs(n_event.event_time - event.event_time) < 2:
                    self.depress_index = index + 1
                    self.depressurize_event_signal.emit(n_event)

        max_time = self.log_reader.events[max(self.press_index, self.depress_index, self.period_index)].event_time
        time = max_time - self.log_reader.events[0].event_time
        self.time_edit.setText(str(ceil(time)))

    def emit_last_event(self):
        index = min(self.press_index, self.depress_index, self.period_index) - 1

        while index - 1 >= 0 and self.log_reader.events[index].event_type == Event.PRESSURE:
            index -= 1

        event = self.log_reader.events[index]

        if index < 0:
            return

        if event.event_type == Event.PRESSURIZE:
            self.press_index = index
            self.pressurize_event_signal.emit(event)

        elif event.event_type == Event.DEPRESSURIZE:
            self.depress_index = index
            self.depressurize_event_signal.emit(event)
            # Also emit matching PERIOD
            if index - 1 >= 0:
                n_event = self.log_reader.events[index - 1]
                if n_event.event_type == Event.PERIOD and abs(n_event.event_time - event.event_time) < 2:
                    self.period_index = index - 1
                    self.period_event_signal.emit(n_event)

        elif event.event_type == Event.PERIOD:
            self.period_index = index
            self.period_event_signal.emit(event)
            # Also emit matching DEPRESSURIZE
            if index - 1 >= 0:
                n_event = self.log_reader.events[index - 1]
                if n_event.event_type == Event.DEPRESSURIZE and abs(n_event.event_time - event.event_time) < 1:
                    self.depress_index = index - 1
                    self.depressurize_event_signal.emit(n_event)

        max_time = self.log_reader.events[max(self.press_index, self.depress_index, self.period_index)].event_time
        time = max_time - self.log_reader.events[0].event_time
        self.time_edit.setText(str(ceil(time)))

    def reset(self):
        self.filename_label.setText("")
        self.time_edit.setText("")
        self.log_coefficients_signal.emit(None)
        self.filename = None

        self.next_button.setEnabled(False)
        self.last_button.setEnabled(False)
        self.edit_button.setEnabled(False)
        self.time_edit.setEnabled(False)

    def set_logging(self, connected):
        self.currently_logging = connected
        self.current_button.setEnabled(connected)
        self.new_log_button.setEnabled(connected)

    def new_log(self):
        if self.currently_logging:
            self.log_reader.logger.new_log_file(self.log_reader.logger.current_path)



import importlib
import subprocess
import os
import platform

# Conditionally import `pwd` only on Unix-based systems
if platform.system() == "Linux":
    import pwd


# Adds permissions for the user to access the USB device
def setup_udev_rules():
    if platform.system() != "Linux":
        raise RuntimeError("This script must be run on Linux systems.")

    with importlib.resources.path('icarus_v2.resources.setup', 'install_udev_rules.sh') as script_path:
        try:
            # Run the script with 'pkexec'
            current_user = pwd.getpwuid(os.getuid()).pw_name
            command = ['pkexec', '--user', current_user, str(script_path)]  # Convert PosixPath to string
            result = subprocess.run(command, stderr=subprocess.PIPE)
            if result.returncode != 0:
                raise RuntimeError(f"Command failed: {result.stderr.decode().strip()}")
        except Exception as e:
            raise RuntimeError(f"Command execution failed: {str(e)}") from e



import lzma
import pickle
import os
from datetime import datetime
from PySide6.QtCore import QStandardPaths
from icarus_v2.backend.configuration_manager import ConfigurationManager


MAX_PRESSURE_INTERVAL = 60


# Logs files to logs/temp or logs/experiment depending on bit 4.
# Files are deleted if no events are logged.
class Logger:
    def __init__(self, is_raw=False) -> None:
        self.file = None
        self.filename = None
        self.current_path = None
        self.event_count = None
        self.last_pressure_update = None
        self.is_raw = is_raw

    def new_log_file(self, temporary=True):
        self.close()

        self.current_path = temporary
        base_dir = os.path.join(QStandardPaths.writableLocation(QStandardPaths.AppDataLocation), 'logs')
        if temporary:
            log_path = os.path.join(base_dir, 'temp')
        else:
            log_path = os.path.join(base_dir, "experiment")

        if not os.path.exists(log_path):
            os.makedirs(log_path)

        current_datetime = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        name = f"{'raw_' if self.is_raw else ''}log_{current_datetime}.xz"
        self.filename = os.path.join(log_path, name)
        self.file = lzma.open(self.filename, "ab")  # Opening file in append binary mode with LZMA compression
        self.event_count = 0
        self.last_pressure_update = None

    def log_event(self, event):
        if self.is_raw:
            raise RuntimeError("Error: Adding a processed event to a raw log.")

        if event.event_type == event.PRESSURE:
            if self.last_pressure_update is None or event.event_time - self.last_pressure_update < MAX_PRESSURE_INTERVAL:
                return

        if event.event_type == event.PRESSURE or event.event_type == event.DEPRESSURIZE:
            self.last_pressure_update = event.event_time

        event_dict = {
            'event_type': event.event_type,
            'data': event.data,
            'event_time': event.event_time,
            'event_index': event.event_index,
            'step_time': event.step_time
        }   
        self.log_raw(event_dict)

    def log_error(self, event):
        return

        #Should take in an instance of either sentry_error or sentry_warning
        if self.is_raw:
            raise RuntimeError("Error: Adding a processed error to a raw log.")

        event_dict = {
            'class_type' : type(event),     #This serves as a flag so that when the log is read it will know if it is a warning or error
            'error_type' : event.error_type,
            'event_time': event.time,
            'data': event.info
        }   

        self.log_raw(event_dict)

    def log_raw(self, data):
        self.event_count += 1
        pickle.dump(data, self.file, protocol=pickle.HIGHEST_PROTOCOL)

    def flush(self):
        self.file.close()
        self.file = lzma.open(self.filename, "ab")

    def close(self):
        if self.file is None:
            return

        if not self.is_raw:
            config_manager = ConfigurationManager()
            settings = {"plotting_coefficients": config_manager.get_settings("plotting_coefficients")}
            pickle.dump(settings, self.file, protocol=pickle.HIGHEST_PROTOCOL)

        self.file.close()
        self.file = None

        if self.event_count == 0:
            try:
                os.remove(self.filename)
            except FileNotFoundError:
                pass

        self.filename = None



from PySide6.QtCore import QThread
import traceback
from icarus_v2.backend.event import Event

class EventHandler(QThread):
    def __init__(self, loader, signal, sample_rate, update_rate) -> None:
        super().__init__()
        self.reader = loader.new_reader()
        self.signal = signal
        self.sample_rate = sample_rate
        self.update_rate = update_rate
        self.running = False


    # Loops to transmit data if an event occurs
    def run(self):
        self.running = True
        while self.running:
            data_to_get = int(self.sample_rate / self.update_rate)
            try:
                data, buffer_index = self.reader.read(size=data_to_get, timeout=1)
            except TimeoutError:
                self.running = False
                break
            try:
                event, chunk_index = self.detect_event(data)
            except RuntimeWarning:
                # Case where 2 events occur in same chunk
                traceback.print_exc()
                event, chunk_index = False, -1

            # buffer_index is the index that the chunk started in in the buffer
            # chunk index is the index that the event started in in that chunk
            event_index = buffer_index + chunk_index

            # If an event occurs, transmit data to plot
            if event:
                event_data, event_start = self.handle_event(event_index)
                if event_data is not None:
                    new_event = Event(self.event_type, event_data, event_start)
                    self.signal.emit(new_event)


    # Placeholder. 
    # Data: one chunk from the reader
    # Returns whether an event occurs and the index of the event
    def detect_event(self, data):
        return False, -1


    # Placeholder.
    # event_index: absolute index of event in buffer
    # Returns appropriate data and index of the event in the new data
    def handle_event(self, event_index):
        return None, -1


    # Return a range of data around the event occuring at event_index
    # Event coordinate is a tuple of the index in the buffer and in the chunk where the event occured
    def get_event_data(self, event_index):
        before, after = self.event_report_range # Amount of data in ms to report around the event

        # Calculate start and end of data to get
        sample_rate_kHz = float(self.sample_rate) / 1000
        start = event_index + int(before * sample_rate_kHz)
        end = event_index + int(after * sample_rate_kHz)

        # Case where more data was requested than is stored in the buffer
        # Safety factor of 0.9 to avoid race conditions
        if end - start > 0.9 * self.reader.buffer.capacity:
            start = int(end - 0.9 * self.reader.buffer.capacity)

        # Get data
        try:
            data = self.reader.retrieve_range(start, end, timeout=2)
        except TimeoutError:
            return None

        return data


    def quit(self):
        self.running = False
